import re

from agilerl.data.language_environment import (
    Language_Environment,
    Language_Observation,
    Policy,
    interact_environment,
)
from agilerl.data.rl_data import (
    ConstantTokenReward,
    DataPoint,
    Iterable_RL_Dataset,
    List_RL_Dataset,
    SpecifiedTokenReward,
    TokenReward,
)
from agilerl.data.tokenizer import Tokenizer
from agilerl.data.torch_datasets import GeneralDataset, GeneralIterDataset
from agilerl.utils.cache import Cache


class WordleTokenizer(Tokenizer):
    def __init__(self):
        self.special_vocab = [
            "<g>",
            "<b>",
            "<y>",
            "<|pad|>",
            "</a>",
            "</s>",
            "<s>",
            "<a>",
            "</eod>",
        ]
        self.vocab = list("abcdefghijklmnopqrstuvwxyz") + self.special_vocab
        self.t2i = {w: i for i, w in enumerate(self.vocab)}
        super().__init__(
            self.token_to_id("<|pad|>"),
            self.token_to_id("</s>"),
            self.token_to_id("</a>"),
            self.token_to_id("<s>"),
            self.token_to_id("<a>"),
            self.token_to_id("</eod>"),
        )

    def encode(self, str_, **kwargs):
        if isinstance(str_, str):
            special_idxs = []
            for special_char in self.special_vocab:
                special_idxs += list(
                    map(
                        lambda x: (x.start(), x.end(), self.token_to_id(special_char)),
                        re.finditer(re.escape(special_char), str_),
                    )
                )
            special_idxs.sort(key=lambda x: x[0])
            tokens = []
            curr = 0
            for s, e, tok in special_idxs:
                tokens.extend([self.token_to_id(c) for c in str_[curr:s]])
                tokens.append(tok)
                curr = e
            tokens.extend([self.token_to_id(c) for c in str_[curr:]])
            return tokens, [int(t != self.pad_token_id) for t in tokens]
        elif isinstance(str_, list):
            tokens, pads = zip(*[self.encode(item) for item in str_])
            max_len = max(map(len, tokens))
            return [
                list(item) + ([self.pad_token_id] * (max_len - len(item)))
                for item in tokens
            ], [list(item) + ([0] * (max_len - len(item))) for item in pads]
        else:
            raise ValueError("str_ must be a string or a list of strings")

    def decode(self, tokens, **kwargs):
        if len(tokens) == 0:
            return ""
        if not isinstance(tokens[0], list):
            return "".join([self.id_to_token(item) for item in tokens])
        elif isinstance(tokens[0], list):
            return [self.decode(item) for item in tokens]
        else:
            raise ValueError("tokens must be a list of ints or a list of lists of ints")

    def num_tokens(self):
        return len(self.vocab)

    def id_to_token(self, id_):
        return self.vocab[id_]

    def token_to_id(self, token):
        return self.t2i[token]

    def get_vocab(self):
        return self.vocab


def test_language_observation_base_class():
    Language_Observation.__abstractmethods__ = set()
    lang_obs = Language_Observation()
    lang_obs.to_sequence()
    lang_obs.__str__()
    metadata = lang_obs.metadata()

    assert metadata is None


def test_language_environment_base_class():
    Language_Environment.__abstractmethods__ = set()

    lang_env = Language_Environment()
    lang_env.step(None)
    lang_env.reset()
    lang_env.is_terminal()

    assert True


def test_policy_base_class():
    Policy.__abstractmethods__ = set()

    policy = Policy()
    policy.act(None)
    policy.train()
    policy.eval()

    assert isinstance(policy.cache, Cache)


def test_interact_environment():
    class LangEnv(Language_Environment):
        def __init__(self):
            super().__init__()
            self.term = False

        def step(self, action):
            self.term = True
            return 1, 1, 1

        def reset(self):
            self.term = False
            return 0

        def is_terminal(self):
            return self.term

    class Pol(Policy):
        def act(self, obs):
            return 1

        def train(self):
            return 0

        def eval(self):
            return False

    env = LangEnv()
    policy = Pol()
    obs = None

    obs, obs_sequence = interact_environment(env, policy, obs)

    assert isinstance(obs, int)
    assert isinstance(obs_sequence, list)
    assert len(obs_sequence) == 2


def test_tokenreward_base_class():
    TokenReward.__abstractmethods__ = set()

    tok_rew = TokenReward()
    tok_rew.get_token_reward(None)

    assert True


def test_constanttokenreward_base_class():
    tokens = [1, 2, 3]
    tok_rew = ConstantTokenReward(1)
    reward = tok_rew.get_token_reward(tokens)

    assert tok_rew.c == 1
    assert reward == [1, 1]


def test_specifiedtokenreward_base_class():
    token_data = {0: 10.0}
    scale = 1.0
    shift = 2.0
    tokens = [0, 0, 1, 2]
    tok_rew = SpecifiedTokenReward(token_data, scale, shift)
    reward = tok_rew.get_token_reward(tokens)

    assert tok_rew.token_data == token_data
    assert tok_rew.scale == scale
    assert tok_rew.shift == shift
    assert reward == [12.0, 2.0, 2.0]


def test_list_rl_dataset_base_class():
    List_RL_Dataset.__abstractmethods__ = set()
    GeneralDataset.__abstractmethods__ = set()

    class LangObs(Language_Observation):
        def to_sequence(self, **kwargs):
            return [("test", 1), ("test", None)], True

    lang_obs = LangObs()
    tokenizer = WordleTokenizer()
    token_reward = ConstantTokenReward(1)
    meta = None

    rl_ds = List_RL_Dataset(tokenizer, token_reward, 10)
    rl_ds.get_item(0)
    rl_ds.size()

    dp = DataPoint.from_obs(lang_obs, tokenizer, token_reward, meta)
    _ = DataPoint.from_obs(lang_obs, tokenizer, token_reward, 1)

    reward = dp.get_token_reward(lang_obs, tokenizer, token_reward)

    ds = GeneralDataset(rl_ds, None)
    len_ds = ds.__len__()
    item_ds = ds.__getitem__(0)
    collate_ds = ds.collate([dp])
    collate_simple_ds = ds.collate_simple(None)

    assert isinstance(reward, list)
    assert len_ds is None
    assert item_ds is None
    assert isinstance(collate_ds, dict)
    assert collate_simple_ds is None


def test_dp_from_obs():
    class LangObs(Language_Observation):
        def to_sequence(self, **kwargs):
            return [("test", None)], True

        def metadata(self):
            return {"test": 1}

    lang_obs = LangObs()
    tokenizer = WordleTokenizer()
    token_reward = ConstantTokenReward(1)
    meta = {"test": 1}

    dp = DataPoint.from_obs(lang_obs, tokenizer, token_reward, meta)

    assert isinstance(dp, DataPoint)

    class LangObs(Language_Observation):
        def to_sequence(self, **kwargs):
            return [("test", None)], True

        def metadata(self):
            return {"test": 1}

    lang_obs = LangObs()
    tokenizer = WordleTokenizer()
    token_reward = ConstantTokenReward(1)
    meta = None

    dp = DataPoint.from_obs(lang_obs, tokenizer, token_reward, meta)

    assert isinstance(dp, DataPoint)


def test_iterable_rl_dataset_base_class():
    Iterable_RL_Dataset.__abstractmethods__ = set()
    GeneralIterDataset.__abstractmethods__ = set()

    class LangObs(Language_Observation):
        def to_sequence(self, **kwargs):
            return [("test", 1), ("test", 2), ("test", None)], True

    lang_obs = LangObs()
    tokenizer = WordleTokenizer()
    token_reward = ConstantTokenReward(1)
    meta = None

    iter_rl_ds = Iterable_RL_Dataset(tokenizer, token_reward, 10)
    iter_rl_ds.sample_item()

    dp = DataPoint.from_obs(lang_obs, tokenizer, token_reward, meta)
    _ = DataPoint.from_obs(lang_obs, tokenizer, token_reward, 1)

    reward = dp.get_token_reward(lang_obs, tokenizer, token_reward)

    ds = GeneralIterDataset(iter_rl_ds, None)
    iter_ds = ds.__iter__()
    next_ds = ds.__next__()
    collate_ds = ds.collate([dp])
    collate_simple_ds = ds.collate_simple(None)

    assert isinstance(reward, list)
    assert iter_ds == ds
    assert next_ds is None
    assert isinstance(collate_ds, dict)
    assert collate_simple_ds is None


def test_tokenizer_base_class():
    Tokenizer.__abstractmethods__ = set()

    tok = Tokenizer(0, 1, 2, 3, 4, 5)
    tok.encode("test")
    tok.decode([0, 1, 2])
    tok.num_tokens()
    tok.id_to_token(0)
    tok.token_to_id("a")
    tok.get_vocab()

    assert tok.pad_token_id == 0
    assert tok.eos_token_id == 1
    assert tok.eoa_token_id == 2
    assert tok.bos_token_id == 3
    assert tok.boa_token_id == 4
    assert tok.eod_token_id == 5



import pytest
import torch.nn as nn

from agilerl.modules.dummy import DummyEvolvable


def test_dummy_evolvable_from_module():
    module = DummyEvolvable(module=nn.Linear(10, 10), device="cpu")
    assert module.device == "cpu"
    assert module.module.weight.shape == (10, 10)
    assert module.module.bias.shape == (10,)


def test_dummy_evolvable_from_module_fn():
    module = DummyEvolvable(module_fn=lambda: nn.Linear(10, 10), device="cpu")
    assert module.device == "cpu"
    assert module.module.weight.shape == (10, 10)
    assert module.module.bias.shape == (10,)


def test_dummy_evolvable_raises():
    with pytest.raises(ValueError) as e:
        DummyEvolvable(module_fn=None, module=None, device="cpu")
    assert "Either module or module_fn must be provided." in str(e.value)


def test_dummy_evolvable_from_module_fn_and_kwargs():
    module = DummyEvolvable(
        module_fn=lambda: nn.Linear(10, 10), module_kwargs={}, device="cpu"
    )
    assert module.device == "cpu"
    assert module.module.weight.shape == (10, 10)



import copy
import logging
import warnings
from collections import OrderedDict
from functools import wraps
from typing import Any, Callable, Optional, TypeVar, Union

import fastrand
import numpy as np
import torch
import torch.nn as nn
from accelerate import Accelerator

from agilerl.algorithms import NeuralTS, NeuralUCB
from agilerl.algorithms.core import (
    EvolvableAlgorithm,
    LLMAlgorithm,
    MultiAgentRLAlgorithm,
    RLAlgorithm,
)
from agilerl.modules import EvolvableModule, ModuleDict
from agilerl.typing import (
    EvolvableNetworkType,
    MutationMethod,
    MutationReturnType,
)
from agilerl.utils.algo_utils import remove_compile_prefix
from agilerl.utils.evolvable_networks import compile_model
from agilerl.wrappers.agent import AgentWrapper

IndividualType = TypeVar("IndividualType", bound=EvolvableAlgorithm)
MutationsType = TypeVar("MutationsType", bound="Mutations")
PopulationType = list[IndividualType]
BanditAlgorithm = Union[NeuralUCB, NeuralTS]

torch._dynamo.config.cache_size_limit = 64
torch._logging.set_logs(dynamo=logging.FATAL)


def set_global_seed(seed: Optional[int]) -> None:
    """Set the global seed for random number generators.

    :param seed: Random seed for repeatability
    :type seed: int
    """
    if seed is None:
        return

    np.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)
    fastrand.pcg32_seed(seed)


def get_offspring_eval_modules(
    individual: IndividualType,
) -> tuple[dict[str, EvolvableNetworkType], dict[str, EvolvableNetworkType]]:
    """Get the offsprings of all of the evaluation modules in the individual.

    :param individual: The individual to inspect
    :type individual: EvolvableAlgorithm

    :return: Tuple of offspring policy and the rest of the evaluation modules
    :rtype: tuple[dict[str, NetworkType], dict[str, NetworkType]]
    """
    registry = individual.registry

    offspring_modules = {}
    offspring_policy = {}
    for group in registry.groups:
        eval_module: EvolvableNetworkType = getattr(individual, group.eval_network)

        # Clone the offspring prior to applying mutations
        offspring = eval_module.clone()
        if group.policy:
            offspring_policy[group.eval_network] = offspring
        else:
            offspring_modules[group.eval_network] = offspring

    return offspring_policy, offspring_modules


def get_exp_layer(offspring: EvolvableModule) -> nn.Module:
    """Get the output layer of different types of offsprings for bandit algorithms.
    Returns None if algorithm is not a bandit algorithm.

    :param offspring: The offspring to inspect
    :type offspring: EvolvableModule

    :return: The output layer of the offspring
    :rtype: nn.Module
    """
    if isinstance(offspring, EvolvableModule):
        exp_layer = offspring.get_output_dense()
    else:
        raise ValueError(
            f"Bandit algorithm architecture {type(offspring)} not supported."
        )

    return exp_layer


def reinit_shared_networks(mutation_func=None):
    """Decorator to reinitialize shared networks after architecture and parameter mutations.

    :param mutation_func: The mutation function to decorate
    :type mutation_func: Optional[Callable[[IndividualType], IndividualType]]
    :return: The decorated mutation function or decorator
    :rtype: Callable
    """

    def decorator(func: MutationMethod) -> Callable:
        @wraps(func)
        def wrapper(self: MutationsType, individual: IndividualType) -> IndividualType:
            # Call the original mutation function
            individual = func(self, individual)

            torch._dynamo.reset()  # NOTE: Should we do this?

            # Only proceed if mutation was actually applied
            if individual.mut == "None":
                return individual

            # Recompile individual if necessary
            compiled_model = individual.torch_compiler is not None
            if compiled_model:
                # Set dynamo config before recompilation to avoid guard failures
                torch._dynamo.config.force_parameter_static_shapes = False
                individual.recompile()

            # Reinitialize shared networks to mutated evaluation networks
            for net_group in individual.registry.groups:
                if net_group.shared_networks is not None:
                    for shared_name in net_group.shared_networks:
                        eval_offspring: EvolvableNetworkType = getattr(
                            individual, net_group.eval_network
                        )
                        # Reinitialize shared with frozen weights due to
                        # potential mutation in architecture
                        ind_shared = self._reinit_from_mutated(
                            eval_offspring,
                            remove_prefix=compiled_model,
                        )
                        if self.accelerator is None:
                            ind_shared = ind_shared.to(self.device)

                        if compiled_model:
                            torch._dynamo.config.force_parameter_static_shapes = False
                            ind_shared = compile_model(
                                ind_shared, individual.torch_compiler
                            )

                        setattr(individual, shared_name, ind_shared)

            return individual

        return wrapper

    return decorator(mutation_func)


class Mutations:
    """Allows performing mutations on a population of :class:`EvolvableAlgorithm <agilerl.algorithms.core.EvolvableAlgorithm>` agents. Calling
    :func:`Mutations.mutation() <agilerl.hpo.mutation.Mutations.mutation>` on a population of agents will return a mutated population of agents.
    The type of mutation applied to each agent is sampled randomly from the probabilities given by the user. The supported types of mutations that
    can be applied to an agent are:

    * No mutation
    * Network architecture mutation - adding layers or nodes. Trained weights are reused and new weights are initialized randomly.
    * Network parameters mutation - mutating weights with Gaussian noise.
    * Network activation layer mutation - change of activation layer.
    * RL algorithm mutation - mutation of learning hyperparameter, (e.g. learning rate or batch size).

    See :ref:`evo_hyperparam_opt` for more details.

    :param no_mutation: Relative probability of no mutation
    :type no_mutation: float
    :param architecture: Relative probability of architecture mutation
    :type architecture: float
    :param new_layer_prob: Relative probability of new layer mutation (type of architecture mutation)
    :type new_layer_prob: float
    :param parameters: Relative probability of network parameters mutation
    :type parameters: float
    :param activation: Relative probability of activation layer mutation
    :type activation: float
    :param rl_hp: Relative probability of learning hyperparameter mutation
    :type rl_hp: float
    :param rl_hp_selection: Learning hyperparameter mutations to choose from
    :type rl_hp_selection: list[str]
    :param mutation_sd: Mutation strength
    :type mutation_sd: float
    :param activation_selection: Activation functions to choose from, defaults to ["ReLU", "ELU", "GELU"]
    :type activation_selection: list[str], optional
    :param mutate_elite: Mutate elite member of population, defaults to True
    :type mutate_elite: bool, optional
    :param rand_seed: Random seed for repeatability, defaults to None
    :type rand_seed: int, optional
    :param device: Device for accelerated computing, 'cpu' or 'cuda', defaults to 'cpu'
    :type device: str, optional
    :param accelerator: Accelerator for distributed computing, defaults to None
    :type accelerator: accelerate.Accelerator(), optional
    """

    def __init__(
        self,
        no_mutation: float,
        architecture: float,
        new_layer_prob: float,
        parameters: float,
        activation: float,
        rl_hp: float,
        mutation_sd: float = 0.1,
        activation_selection: list[str] = ["ReLU", "ELU", "GELU"],
        mutate_elite: bool = True,
        rand_seed: Optional[int] = None,
        device: str = "cpu",
        accelerator: Optional[Accelerator] = None,
    ):
        assert isinstance(
            no_mutation, (float, int)
        ), "Probability of no mutation must be a float or integer."
        assert (
            no_mutation >= 0
        ), "Probability of no mutation must be greater than or equal to zero."
        assert isinstance(
            architecture, (float, int)
        ), "Probability of architecture mutation must be a float or integer."
        assert (
            architecture >= 0
        ), "Probability of architecture mutation must be greater than or equal to zero."
        assert isinstance(
            new_layer_prob, (float, int)
        ), "Probability of new layer architecture mutation must be a float or integer."
        assert (
            1 >= new_layer_prob >= 0
        ), "Probability of new layer architecture mutation must be between zero and one (inclusive)."
        assert isinstance(
            parameters, (float, int)
        ), "Probability of parameters mutation must be a float or integer."
        assert (
            parameters >= 0
        ), "Probability of parameters mutation must be greater than or equal to zero."
        assert isinstance(
            activation, (float, int)
        ), "Probability of activation mutation must be a float or integer."
        assert (
            activation >= 0
        ), "Probability of activation mutation must be greater than or equal to zero."
        assert isinstance(
            rl_hp, (float, int)
        ), "Probability of reinforcement learning hyperparameter mutation must be a float or integer."
        assert (
            rl_hp >= 0
        ), "Probability of reinforcement learning hyperparameter mutation must be greater than or equal to zero."
        assert (
            mutation_sd >= 0
        ), "Mutation strength must be greater than or equal to zero."
        assert isinstance(
            mutation_sd, (float, int)
        ), "Mutation strength must be a float or integer."
        assert isinstance(
            mutate_elite, bool
        ), "Mutate elite must be boolean value True or False."
        assert (
            isinstance(rand_seed, int) or rand_seed is None
        ), "Random seed must be an integer or None."
        if isinstance(rand_seed, int):
            assert rand_seed >= 0, "Random seed must be greater than or equal to zero."

        # Random seed for repeatability
        set_global_seed(rand_seed)
        self.rng = np.random.default_rng(rand_seed)

        # Relative probabilities of mutation
        self.no_mut = no_mutation  # No mutation
        self.architecture_mut = architecture  # Architecture mutation

        # New layer mutation (type of architecture mutation)
        self.new_layer_prob = new_layer_prob
        self.parameters_mut = parameters  # Network parameters mutation
        self.activation_mut = activation  # Activation layer mutation
        self.rl_hp_mut = rl_hp  # Learning HP mutation
        self.activation_selection = activation_selection  # Learning HPs to choose from
        self.mutation_sd = mutation_sd  # Mutation strength
        self.mutate_elite = mutate_elite
        self.device = device
        self.accelerator = accelerator

        self.pretraining_mut_options, self.pretraining_mut_proba = (
            self._get_mutations_options(pretraining=True)
        )
        self.mut_options, self.mut_proba = self._get_mutations_options()

    def mutation(
        self, population: PopulationType, pre_training_mut: bool = False
    ) -> PopulationType:
        """Returns a mutated population of agents. See :ref:`evo_hyperparam_opt` for more details.

        :param population: Population of agents
        :type population: list[EvolvableAlgorithm]
        :param pre_training_mut: Boolean flag indicating if the mutation is before the training loop
        :type pre_training_mut: bool, optional

        :return: Mutated population
        :rtype: list[EvolvableAlgorithm]
        """
        # Create lists of possible mutation functions and their respective relative probabilities
        mutation_options = (
            self.pretraining_mut_options if pre_training_mut else self.mut_options
        )
        mutation_proba = (
            self.pretraining_mut_proba if pre_training_mut else self.mut_proba
        )

        # Randomly choose mutation for each agent in population from options with
        # relative probabilities
        mutation_choice: list[MutationMethod] = self.rng.choice(
            mutation_options, len(population), p=mutation_proba
        )

        # If not mutating elite member of population (first in list from tournament selection),
        # set this as the first mutation choice
        if not self.mutate_elite:
            mutation_choice[0] = self.no_mutation

        mutated_population = []
        for mutation, individual in zip(mutation_choice, population):
            wrapped_ind = isinstance(individual, AgentWrapper)
            agent = individual.agent if wrapped_ind else individual

            agent = mutation(agent)  # Call sampled mutation for individual
            agent.mutation_hook()  # Call hooks specified by user

            if wrapped_ind:
                individual.agent = agent
            else:
                individual = agent

            mutated_population.append(individual)

        return mutated_population

    def no_mutation(self, individual: IndividualType):
        """Returns individual from population without mutation.

        :param individual: Individual agent from population
        :type individual:
        """
        individual.mut = "None"  # No mutation
        return individual

    @reinit_shared_networks
    def architecture_mutate(self, individual: IndividualType) -> IndividualType:
        """Performs a random mutation to the architecture of the policy network of an agent. The way in
        which we apply an architecture mutation to single and multi-agent RL algorithms inherently differs
        given the nested nature of the networks in the latter.

        * **Single-agent:** A mutation method is sampled from the policy network and then applied to the rest of the evaluation
          modules (e.g. critics). This can be done generally because all of the networks in a single-agent algorithm share the same
          architecture (given there is only one observation space).

        * **Multi-agent:** A sub-agent is sampled to perform the mutation on for the policy. We then iterate over the rest of the
          sub-agent policies and perform the same mutation if they share the same observation space. For the rest of the evaluation
          networks (e.g. critics) there is a possibility they are centralized, in which case their underlying architecture
          will differ from the policy and therefore the mutation methods won't exactly match. In such cases, we try to find an analogous
          mutation method to apply.

        .. note::
            This is currently not supported for :class:`LLMAlgorithm <agilerl.algorithms.core.LLMAlgorithm>` agents.

        :param individual: Individual agent from population
        :type individual: RLAlgorithm or MultiAgentRLAlgorithm

        :return: Individual from population with network architecture mutation
        :rtype: RLAlgorithm or MultiAgentRLAlgorithm
        """
        if isinstance(individual, RLAlgorithm):
            individual = self._architecture_mutate_single(individual)
        elif isinstance(individual, MultiAgentRLAlgorithm):
            individual = self._architecture_mutate_multi(individual)
        else:
            raise MutationError(
                f"Architecture mutations are not supported for {individual.__class__.__name__}. "
                "Please make sure your algorithm inherits from 'RLAlgorithm' or 'MultiAgentRLAlgorithm'."
            )

        return individual

    def rl_hyperparam_mutation(self, individual: IndividualType) -> IndividualType:
        """Performs a random mutation of a learning hyperparameter of an agent. To do this, sample a hyperparameter from those
        specified through the :class:`HyperparameterConfig <agilerl.algorithms.core.registry.HyperparameterConfig>`
        passed during initialization of the agent. The hyperparameter is then mutated and the optimizer is reinitialized if the
        learning rate has been mutated.

        :param individual: Individual agent from population
        :type individual: EvolvableAlgorithm

        :return: Individual from population with RL hyperparameter mutation
        :rtype: EvolvableAlgorithm
        """
        # Randomly sample hyperparameter to mutate from the passed configuration
        hp_config = individual.registry.hp_config
        if not hp_config:
            individual.mut = "None"
            return individual

        mutate_attr, mutate_param = hp_config.sample()

        if mutate_param.value is None:
            mutate_param.value = getattr(individual, mutate_attr)

        # Randomly grow or shrink hyperparameters by specified factors
        new_value = mutate_param.mutate()

        setattr(individual, mutate_attr, new_value)

        # Reinitialize optimizer if mutated learning rate
        if mutate_attr in individual.get_lr_names():
            optimizer_configs = individual.registry.optimizers
            to_reinit = [
                opt_config
                for opt_config in optimizer_configs
                if mutate_attr == opt_config.lr
            ][0]

            individual.reinit_optimizers(optimizer=to_reinit)

        individual.mut = mutate_attr
        return individual

    # TODO: Activation mutations should really be integrated as architecture mutations
    @reinit_shared_networks
    def activation_mutation(self, individual: IndividualType) -> IndividualType:
        """Performs a random mutation of the activation layer of the evaluation networks of an agent.

        .. note::
            This is currently not supported for :class:`LLMAlgorithm <agilerl.algorithms.core.LLMAlgorithm>` agents.

        :param individual: Individual agent from population
        :type individual: RLAlgorithm or MultiAgentRLAlgorithm

        :return: Individual from population with activation layer mutation
        :rtype: RLAlgorithm or MultiAgentRLAlgorithm
        """
        # Needs to stay constant for policy gradient methods
        # NOTE: Could set up an algorithm registry to make algo checks more robust
        # OR perform activation mutations within evolvable modules directly and disable
        # on an algorithm basis
        if individual.algo in ["PPO", "DDPG", "TD3", "IPPO", "MADDPG", "MATD3", "GRPO"]:
            warnings.warn(
                f"Activation mutations are not supported for {individual.algo}."
            )
            individual.mut = "None"
            return individual

        # Mutate network activation layer
        registry = individual.registry
        no_activation = False
        for network_group in registry.groups:
            eval_module: EvolvableNetworkType = getattr(
                individual, network_group.eval_network
            )

            if eval_module.activation is None:
                no_activation = True
            else:
                eval_module = self._permutate_activation(eval_module)

            if no_activation:
                warnings.warn(
                    "Found no activation mutation capabilities. We advise setting the probability to "
                    "0.0 to disable activation mutations."
                )
                break

            if self.accelerator is None:
                eval_module = eval_module.to(self.device)

            if isinstance(individual, (NeuralTS, NeuralUCB)):
                individual.exp_layer = get_exp_layer(eval_module)

            setattr(individual, network_group.eval_network, eval_module)

        individual.reinit_optimizers()  # Reinitialize optimizer
        individual.mut = "act" if not no_activation else "None"
        return individual

    def parameter_mutation(self, individual: IndividualType) -> IndividualType:
        """Performs a random mutation to the weights of the policy network of an agent through
        the addition of Gaussian noise.

        .. note::
            This is currently not supported for :class:`LLMAlgorithm <agilerl.algorithms.core.LLMAlgorithm>` agents.

        :param individual: Individual agent from population
        :type individual: RLAlgorithm or MultiAgentRLAlgorithm

        :return: Individual from population with network parameters mutation
        :rtype: RLAlgorithm or MultiAgentRLAlgorithm
        """
        if isinstance(individual, LLMAlgorithm):
            warnings.warn(
                "Parameter mutations are not supported for LLM algorithms. Skipping mutation."
            )
            individual.mut = "None"
            return individual

        registry = individual.registry

        # We only apply parameter mutations to the evaluation policy network
        # (i.e. the network used to select actions)
        policy_group = registry.policy(return_group=True)
        offspring_policy: EvolvableNetworkType = getattr(
            individual, policy_group.eval_network
        )
        if isinstance(offspring_policy, ModuleDict):
            for agent_id, module in offspring_policy.items():
                offspring_policy[agent_id] = self._gaussian_parameter_mutation(module)
        else:
            offspring_policy = self._gaussian_parameter_mutation(offspring_policy)

        self._to_device_and_set_individual(
            individual, policy_group.eval_network, offspring_policy
        )

        # Load state dicts for shared networks
        if policy_group.shared_networks is not None:
            for shared in policy_group.shared_networks:
                offspring_shared: EvolvableNetworkType = getattr(individual, shared)
                offspring_shared.load_state_dict(
                    offspring_policy.state_dict(), strict=False
                )
                self._to_device_and_set_individual(individual, shared, offspring_shared)

        individual.reinit_optimizers()  # Reinitialize optimizer
        individual.mut = "param"

        return individual

    def _get_mutations_options(
        self, pretraining: bool = False
    ) -> tuple[list[Callable], list[float]]:
        """Get the mutation options and probabilities for the given mutation
        configuration.

        :param pretraining: Boolean flag indicating if the mutation is before the training loop
        :type pretraining: bool
        :return: Mutation functions and their respective relative probabilities
        :rtype: tuple[list[Callable], list[float]]
        """
        # Create lists of possible mutation functions and their
        # respective relative probabilities
        mutation_options = [
            (self.no_mutation, self.no_mut),
            (self.architecture_mutate, self.architecture_mut),
            (self.parameter_mutation, self.parameters_mut),
            (self.activation_mutation, self.activation_mut),
            (self.rl_hyperparam_mutation, self.rl_hp_mut),
        ]

        if pretraining:
            mutation_options[0] = (self.no_mutation, 0)

        mutation_options = [(func, prob) for func, prob in mutation_options if prob > 0]

        # This will really only happen when pretraining is True and user has set
        # all mutation probabilities to zero, hence we apply no mutation
        if len(mutation_options) == 0:
            mutation_options = [(self.no_mutation, 1)]

        mutation_funcs, mutation_proba = zip(*mutation_options)
        mutation_proba = np.array(mutation_proba) / np.sum(mutation_proba)
        return mutation_funcs, mutation_proba

    def _to_device_and_set_individual(
        self, individual: IndividualType, name: str, networks: EvolvableNetworkType
    ) -> None:
        """Moves networks to the device and assigns them back to the individual.

        :param individual: The individual to assign the networks to
        :type individual: EvolvableAlgorithm
        :param name: The name of the attribute to assign the networks to
        :type name: str
        :param networks: The networks to move to the device
        :type networks: EvolvableNetworkType
        """
        if self.accelerator is None:
            networks = networks.to(self.device)

        setattr(individual, name, networks)

    def _reinit_module(
        self, module: EvolvableModule, init_dict: dict[str, Any]
    ) -> EvolvableModule:
        """Reinitialize the module with the given initialization dictionary.

        :param module: The module to reinitialize
        :type module: EvolvableModule
        :param init_dict: The initialization dictionary
        :type init_dict: dict[str, Any]

        :return: The reinitialized module
        :rtype: EvolvableModule
        """
        module_orig = (
            module._orig_mod
            if isinstance(module, torch._dynamo.eval_frame.OptimizedModule)
            else module
        )
        return type(module_orig)(**init_dict)

    def _reinit_from_mutated(
        self, offspring: EvolvableNetworkType, remove_prefix: bool = False
    ) -> EvolvableNetworkType:
        """Reinitialize the mutated offspring with their state dictionary.

        :param offspring: The offspring to reinitialize
        :type offspring: NetworkType
        :param remove_prefix: Whether to remove the prefix from the offspring
        :type remove_prefix: bool

        :return: The reinitialized offspring
        :rtype: EvolvableNetworkType
        """
        if isinstance(offspring, ModuleDict):
            reinit_modules: dict[str, EvolvableModule] = OrderedDict()
            for agent_id in offspring:
                nested_offspring: EvolvableModule = offspring[agent_id]
                reinit_modules[agent_id] = self._reinit_module(
                    nested_offspring, nested_offspring.init_dict
                )

            state_dicts = {
                agent_id: nested_offspring.state_dict()
                for agent_id, nested_offspring in offspring.items()
            }
            self._load_state_dicts(reinit_modules, state_dicts, remove_prefix)

            ind_shared = ModuleDict(reinit_modules)
        else:
            ind_shared = self._reinit_module(offspring, offspring.init_dict)
            ind_shared.load_state_dict(offspring.state_dict(), strict=False)

        return ind_shared

    def _load_state_dicts(
        self,
        modules: ModuleDict[EvolvableModule],
        state_dicts: dict[str, dict[str, Any]],
        remove_prefix: bool = False,
    ) -> None:
        """Load the state dictionaries for a multi-agent ModuleDict.

        :param modules: The modules to load the state dictionary into
        :type modules: ModuleDict[EvolvableModule]
        :param state_dicts: The state dictionary to load
        :type state_dicts: dict[str, dict[str, Any]]
        :param remove_prefix: Whether to remove the prefix from the state dictionary
        :type remove_prefix: bool
        """
        for agent_id, module in modules.items():
            state_dict = (
                remove_compile_prefix(state_dicts[agent_id])
                if remove_prefix
                else state_dicts[agent_id]
            )
            module.load_state_dict(state_dict, strict=False)

    def _permutate_activation(self, network: EvolvableModule) -> EvolvableModule:
        """Permutate the activation layer of the network.

        :param network: The network to permutate the activation layer for
        :type network: EvolvableModule

        :return: The network with permutated activation layer
        :rtype: EvolvableModule
        """
        # Function to change network activation layer
        possible_activations = copy.deepcopy(self.activation_selection)
        current_activation = network.activation

        # Remove current activation from options to ensure different new activation layer
        if len(possible_activations) > 1 and current_activation in possible_activations:
            possible_activations.remove(current_activation)

        # Select new activation and modify network
        new_activation = self.rng.choice(possible_activations, size=1)[0]
        network.change_activation(new_activation, output=False)

        return network

    def _gaussian_parameter_mutation(self, network: EvolvableModule) -> EvolvableModule:
        """
        Returns network with mutated weights using a Gaussian distribution.

        :param network: Neural network to mutate.
        :type network: EvolvableModule
        :return: Mutated network.
        :rtype: EvolvableModule
        """
        # Parameters controlling mutation strength and probabilities
        mut_strength = self.mutation_sd
        num_mutation_frac = 0.1
        super_mut_strength = 10
        super_mut_prob = 0.05
        reset_prob = super_mut_prob + 0.05
        mag_limit = 1000000

        model_params: dict[str, torch.Tensor] = network.state_dict()

        # Collect keys corresponding to weight matrices (ignoring normalization / lstm params)
        exclude_keys = ["lstm", "norm"]
        potential_keys = [
            key
            for key in model_params
            if all(exclude_key not in key for exclude_key in exclude_keys)
            and len(model_params[key].shape) == 2
        ]

        # Randomly choose a subset of keys to mutate
        how_many = int(self.rng.integers(1, len(potential_keys) + 1))
        chosen_keys = self.rng.choice(potential_keys, how_many, replace=False)

        for key in chosen_keys:
            W: torch.Tensor = model_params[key]
            num_weights = W.shape[0] * W.shape[1]
            num_mutations = int(np.ceil(num_mutation_frac * num_weights))
            if num_mutations < 1:
                continue

            # Vectorized generation of random indices (for rows and columns)
            rows = self.rng.integers(0, W.shape[0], size=num_mutations)
            cols = self.rng.integers(0, W.shape[1], size=num_mutations)
            rand_vals = self.rng.uniform(0, 1, size=num_mutations)

            # Convert indices and random values to torch tensors on the same device as W
            rows_tensor = torch.tensor(rows, dtype=torch.long, device=W.device)
            cols_tensor = torch.tensor(cols, dtype=torch.long, device=W.device)
            rand_vals_tensor = torch.tensor(rand_vals, dtype=W.dtype, device=W.device)

            # Get current weight values at the selected indices
            current_vals: torch.Tensor = W[rows_tensor, cols_tensor]
            new_vals = current_vals.clone()

            # Create masks for the different mutation types
            mask_super = rand_vals_tensor < super_mut_prob
            mask_reset = (rand_vals_tensor >= super_mut_prob) & (
                rand_vals_tensor < reset_prob
            )
            mask_normal = rand_vals_tensor >= reset_prob

            # Super mutation: add noise with std proportional to the absolute current value times super_mut_strength
            if mask_super.sum() > 0:
                std_super = (super_mut_strength * current_vals[mask_super]).abs()
                noise_super = torch.normal(
                    mean=torch.zeros_like(std_super), std=std_super
                )
                new_vals[mask_super] = current_vals[mask_super] + noise_super

            # Reset mutation: completely reset the weight using N(0, 1)
            if mask_reset.sum() > 0:
                noise_reset = torch.normal(
                    mean=torch.zeros(mask_reset.sum(), device=W.device),
                    std=torch.ones(mask_reset.sum(), device=W.device),
                )
                new_vals[mask_reset] = noise_reset

            # Normal mutation: add noise with std proportional to the absolute current value times mut_strength
            if mask_normal.sum() > 0:
                std_normal = (mut_strength * current_vals[mask_normal]).abs()
                noise_normal = torch.normal(
                    mean=torch.zeros_like(std_normal), std=std_normal
                )
                new_vals[mask_normal] = current_vals[mask_normal] + noise_normal

            # Integrate regularization by clamping all mutated values at once.
            # This is equivalent to your regularize_weight function.
            new_vals = new_vals.clamp(min=-mag_limit, max=mag_limit)

            # Write the mutated, clamped values back to the weight tensor
            W[rows_tensor, cols_tensor] = new_vals
            if self.accelerator is None:
                network = network.to(self.device)

        return network

    def _architecture_mutate_single(self, individual: RLAlgorithm) -> RLAlgorithm:
        """
        Apply an architecture mutation to a single-agent RL algorithm. Since all of the
        networks in a single-agent algorithm share the same architecture (given there is
        only one observation space), we first sample a mutation method from the policy network
        and then apply the same mutation to the rest of the evaluation modules (e.g. critics).
        This is preferred since it reduces variance attributed to evolutionary HPO during training
        and different evaluation networks usually solve tasks of similar complexity and should
        therefore share a similar architecture.

        :param individual: Individual agent from population
        :type individual: RLAlgorithm

        :return: Individual from population with network architecture mutation
        :rtype: RLAlgorithm
        """
        # Get the offspring evaluation modules
        # We first extract and apply a mutation to the policy and then apply
        # the same mutation to the rest of the evaluation modules e.g. critics
        policy, offspring_evals = get_offspring_eval_modules(individual)
        policy_name, policy_offspring = list(policy.items())[0]

        if not policy_offspring.mutation_methods:
            warnings.warn(
                "No mutation methods found for the policy network. Skipping architecture mutation. "
                "We advise setting the probability of architecture mutations to zero when using non-evolvable networks."
            )
            individual.mut = "None"
            return individual

        # Sample mutation method from policy network
        mut_method = policy_offspring.sample_mutation_method(
            self.new_layer_prob, self.rng
        )

        applied_mutation, mut_dict = self._apply_arch_mutation(
            policy_offspring, mut_method
        )
        self._to_device_and_set_individual(individual, policy_name, policy_offspring)

        if isinstance(individual, (NeuralTS, NeuralUCB)):
            old_exp_layer = get_exp_layer(policy_offspring)
            self._reinit_bandit_grads(individual, policy_offspring, old_exp_layer)

        # Apply the same mutation to the rest of the evaluation modules
        for name, offspring in offspring_evals.items():
            if applied_mutation in offspring.mutation_methods:
                self._apply_arch_mutation(offspring, applied_mutation, mut_dict)
                self._to_device_and_set_individual(individual, name, offspring)

        individual.mutation_hook()  # Apply mutation hook
        individual.reinit_optimizers()  # Reinitialize optimizer
        individual.mut = applied_mutation or "None"

        return individual

    def _architecture_mutate_multi(
        self, individual: MultiAgentRLAlgorithm
    ) -> MultiAgentRLAlgorithm:
        """
        Apply an architecture mutation to a multi-agent RL algorithm. Since each agent has its own
        observation space, we can't generally apply the same architecture mutation to all sub-agents.
        Instead, we sample a sub-agent to perform the mutation on for the policy. We then iterate over
        the rest of the sub-agent policies and perform the same mutation if they share the same observation
        space. For the rest of the evaluation networks (e.g. critics) there is a possibility they are
        shared by multiple agents, in which case their underlying architecture will differ from the policy and
        therefore the mutation methods won't exactly match. In this case, we try to find an analogous
        mutation method to apply.

        .. note::
            Since we use `agilerl.modules.ModuleDict` to store multi-agent networks, the available mutation
            methods will have the form ``<agent_id>.<mutation_method>``.

        :param individual: Individual agent from population
        :type individual: MultiAgentRLAlgorithm

        :return: Individual from population with network architecture mutation
        :rtype: MultiAgentRLAlgorithm
        """
        # Get the offspring evaluation modules
        # We first extract and apply a mutation to the policy and then apply
        # the same mutation to the rest of the evaluation modules e.g. critics
        policy, offspring_evals = get_offspring_eval_modules(individual)
        policy_name, policy_offspring = list(policy.items())[0]

        if not policy_offspring.mutation_methods:
            warnings.warn(
                "No mutation methods found for the policy network. Skipping architecture mutation. "
                "We advise setting the probability of architecture mutations to zero when using non-evolvable networks."
            )
            individual.mut = "None"
            return individual

        # Sample mutation method from policy network
        mut_method = policy_offspring.sample_mutation_method(
            self.new_layer_prob, self.rng
        )

        # Apply the sampled method to the policy network (will only apply to one sub-agent)
        applied_mutation, mut_dict = self._apply_arch_mutation(
            policy_offspring, mut_method
        )

        applied_mutations = []
        if applied_mutation is not None:
            split_mutation = applied_mutation.split(".")
            sampled_agent_id = split_mutation[0]
            sampled_mutation = ".".join(split_mutation[1:])
            applied_mutations.append(sampled_agent_id)
        else:
            sampled_agent_id = mut_method.split(".")[0]
            sampled_mutation = None

        # Apply the sampled method to the sub-agents that share the same observation space
        for agent_id, policy in policy_offspring.items():
            if agent_id == sampled_agent_id:
                continue

            # Apply the sampled mutation only if it is available for the current sub-agent
            applied_agent = None
            if sampled_mutation in policy.mutation_methods:
                applied_agent, _ = self._apply_arch_mutation(
                    policy, sampled_mutation, mut_dict
                )

            if applied_agent is not None:
                applied_mutations.append(agent_id)

        self._to_device_and_set_individual(individual, policy_name, policy_offspring)

        # Try to apply an analogous mutation to the rest of the evaluation modules
        for name, offspring_eval in offspring_evals.items():
            # Iterate over the agents in the offspring evaluation module
            for agent_id, agent_eval in offspring_eval.items():
                # Iterate over the the agents whose policies were mutated
                analogous_method = False
                for mutated_agent in applied_mutations:
                    # Don't want to reapply the same method redundantly
                    if (
                        analogous_method
                        and agent_eval.last_mutation_attr == analogous_method
                    ):
                        continue

                    available_methods = agent_eval.mutation_methods

                    # Try to find an analogous mutation method
                    analogous_method = self._find_analogous_mutation(
                        sampled_mutation, available_methods, mutated_agent
                    )

                    if analogous_method is not None:
                        self._apply_arch_mutation(
                            agent_eval, analogous_method, mut_dict
                        )
                    else:
                        raise MutationError(
                            f"Mutation method '{sampled_mutation}' not found in '{agent_eval.__class__.__name__}'. "
                            f"No analogous method found for agent '{agent_id}'. "
                            f"Available methods: {agent_eval.mutation_methods}."
                        )

            self._to_device_and_set_individual(individual, name, offspring_eval)

        individual.mutation_hook()  # Apply mutation hook
        individual.reinit_optimizers()  # Reinitialize optimizer
        individual.mut = sampled_mutation or "None"

        return individual

    def _apply_arch_mutation(
        self,
        network: EvolvableNetworkType,
        mut_method: Optional[str],
        applied_mut_dict: Optional[dict[str, Any]] = None,
    ) -> tuple[Optional[str], MutationReturnType]:
        """Applies the mutation method to networks and returns mutation data if needed.

        :param networks: The networks to apply the mutation to
        :type networks: EvolvableNetworkType
        :param mut_method: The mutation method to apply
        :type mut_method: Optional[str]
        :param applied_mut_dict: The mutation dictionary, defaults to None
        :type applied_mut_dict: Optional[dict[str, Any]], optional

        :return: The mutation method name and the mutation dictionary
        :rtype: tuple[Optional[str], MutationReturnType]
        """
        if not isinstance(network, EvolvableModule):
            raise MutationError(
                f"Can't apply architecture mutation to {network.__class__.__name__} network."
                "Please make sure your network inherits from 'EvolvableModule'."
            )

        applied_mut_dict = applied_mut_dict or {}
        mut_dict = None
        if mut_method is None:
            mut_dict = {}
            network.last_mutation_attr = None
            network.last_mutation = None
        else:
            if mut_method not in network.mutation_methods:
                raise MutationError(
                    f"Mutation method '{mut_method}' not found in '{network.__class__.__name__}'; "
                    f"available methods: \n {network.mutation_methods}."
                )

            mut_dict = getattr(network, mut_method)(**applied_mut_dict)

        mut_dict = mut_dict or {}
        applied_mut = network.last_mutation_attr

        return applied_mut, mut_dict

    # TODO: Can this be implemented as a mutation hook for the bandit algorithms?
    def _reinit_bandit_grads(
        self,
        individual: BanditAlgorithm,
        offspring_actor: EvolvableModule,
        old_exp_layer: nn.Module,
    ) -> None:
        """Reinitialise bandit gradients after architecture mutation.

        :param individual: Individual agent from population
        :type individual: EvolvableAlgorithm
        :param offspring_actor: Offspring actor network
        :type offspring_actor: EvolvableModule
        :param old_exp_layer: Old linear layer
        :type old_exp_layer: nn.Module
        """
        if isinstance(offspring_actor, EvolvableModule):
            exp_layer = offspring_actor.get_output_dense()
        else:
            raise ValueError(
                f"Bandit algorithm architecture {type(offspring_actor)} not supported."
            )

        individual.numel = sum(
            w.numel() for w in exp_layer.parameters() if w.requires_grad
        )
        individual.theta_0 = torch.cat(
            [w.flatten() for w in exp_layer.parameters() if w.requires_grad]
        )

        # create matrix that is copy of sigma inv
        # first go through old params, figure out which to remove, then remove any difference
        # then go through new params, figure out where to add, then add zeros/lambda
        new_sigma_inv = copy.deepcopy(individual.sigma_inv).cpu().numpy()
        old_params = dict(old_exp_layer.named_parameters())
        new_params = dict(exp_layer.named_parameters())

        to_remove = []
        i = 0
        for key, param in old_exp_layer.named_parameters():
            if param.requires_grad:
                old_size = param.numel()
                if key not in new_params.keys():
                    to_remove += list(range(i, i + old_size))
                else:
                    new_size = new_params[key].numel()
                    if new_size < old_size:
                        to_remove += list(range(i + new_size, i + old_size))
                i += old_size

        to_add = []
        i = 0
        for key, param in exp_layer.named_parameters():
            if param.requires_grad:
                new_size = param.numel()
                if key in old_params.keys():
                    old_size = old_params[key].numel()
                    if new_size > old_size:
                        to_add += list(range(i + old_size, i + new_size))
                else:
                    to_add += list(range(i, i + new_size))
                i += new_size

        # Adjust indices to add after deletion
        to_remove = np.array(to_remove)
        to_add = np.array(to_add)
        to_add -= np.sum(to_add[:, np.newaxis] > to_remove, axis=1)
        to_add -= np.arange(len(to_add))

        # Remove elements corresponding to old params
        if len(to_remove) > 0:
            new_sigma_inv = np.delete(
                np.delete(new_sigma_inv, to_remove, 0), to_remove, 1
            )

        # Add new zeros corresponding to new params, make lambda down identity diagonal
        if len(to_add) > 0:
            new_sigma_inv = np.insert(
                np.insert(new_sigma_inv, to_add, 0, 0), to_add, 0, 1
            )
            for i in to_add:
                new_sigma_inv[i, i] = individual.lamb

        individual.exp_layer = exp_layer
        individual.sigma_inv = torch.from_numpy(new_sigma_inv).to(
            individual.device
            if individual.accelerator is None
            else individual.accelerator.device
        )

    def _find_analogous_mutation(
        self, sampled_mutation: str, available_methods: list[str], policy_agent: str
    ) -> Optional[str]:
        """Find an analogous mutation method when exact match is not found.

        Tries to match based on bottom-level method and agent ID.

        :param sampled_mutation: The mutation method that was sampled (e.g., 'encoder.add_channel')
        :type sampled_mutation: str
        :param available_methods: List of available mutation methods
        :type available_methods: list[str]
        :param policy_agent: The agent ID to match (e.g., 'agent_0')
        :type policy_agent: str

        :return: Analogous mutation method if found, None otherwise
        :rtype: Optional[str]
        """
        if not sampled_mutation:
            return None

        elif sampled_mutation in available_methods:
            return sampled_mutation

        sampled_parts = sampled_mutation.split(".")
        bottom_level_method = sampled_parts[-1]

        # Look for methods that:
        # 1. End with the same bottom-level method
        # 2. Contain the policy_agent or 'vector_mlp' as one of the parts
        for method in available_methods:
            method_parts = method.split(".")

            # Check if bottom-level method matches
            if method_parts[-1] == bottom_level_method:
                if policy_agent in method_parts or "vector_mlp" in method_parts:
                    return method

        return None


class MutationError(Exception):
    """Custom exception for mutation errors."""

    pass



from abc import ABC
from collections import OrderedDict
from functools import partial
from typing import Any, Callable, Generic, Optional, TypeVar, Union

import dill
import numpy as np
import torch
from gymnasium import spaces
from tensordict import is_tensor_collection

from agilerl.algorithms import PPO
from agilerl.algorithms.core import (
    EvolvableAlgorithm,
    MultiAgentRLAlgorithm,
    RLAlgorithm,
)
from agilerl.algorithms.core.base import get_checkpoint_dict
from agilerl.typing import (
    ActionReturnType,
    DeviceType,
    ExperiencesType,
    ObservationType,
)
from agilerl.utils.algo_utils import obs_to_tensor, stack_experiences
from agilerl.wrappers.utils import RunningMeanStd

AgentType = TypeVar("AgentType", bound=Union[RLAlgorithm, MultiAgentRLAlgorithm])
MARLObservationType = dict[str, ObservationType]
SelfAgentWrapper = TypeVar("SelfAgentWrapper", bound="AgentWrapper")


class AgentWrapper(ABC, Generic[AgentType]):
    """Base class for all agent wrappers. Agent wrappers are used to apply an
    additional functionality to the ``get_action()`` and ``learn()`` methods of
    an ``EvolvableAlgorithm`` instance.

    :param agent: Agent to be wrapped
    :type agent: AgentType
    """

    wrapped_get_action: Callable
    wrapped_learn: Callable

    def __init__(self, agent: AgentType) -> None:
        self.agent = agent
        self.observation_space = agent.observation_space
        self.action_space = agent.action_space
        self.multi_agent = isinstance(agent, MultiAgentRLAlgorithm)

        # Wrap the agent's methods
        self.wrapped_get_action = agent.get_action
        self.wrapped_learn = agent.learn

        self.agent.get_action = partial(self.get_action)
        self.agent.learn = partial(self.learn)

    @property
    def training(self) -> bool:
        """Returns the training status of the agent.

        :return: Training status of the agent
        :rtype: bool
        """
        return self.agent.training

    @property
    def device(self) -> DeviceType:
        """Returns the device of the agent.

        :return: Device of the agent
        :rtype: DeviceType
        """
        return self.agent.device if not hasattr(self.agent, "rollout_buffer") else "cpu"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.agent})"

    def __getstate__(self) -> dict[str, Any]:
        return {
            "agent": self.agent,
            "observation_space": self.observation_space,
            "action_space": self.action_space,
            "multi_agent": self.multi_agent,
        }

    def __setstate__(self, state: dict[str, Any]) -> None:
        self.agent: AgentType = state["agent"]
        self.observation_space = state["observation_space"]
        self.action_space = state["action_space"]
        self.multi_agent = state["multi_agent"]

        self.wrapped_get_action = self.agent.get_action
        self.wrapped_learn = self.agent.learn

        self.agent.get_action = partial(self.get_action)
        self.agent.learn = partial(self.learn)

    def __getattr__(self, name: str) -> Any:
        """Get attribute of the wrapper.

        :param name: The name of the attribute.
        :type name: str
        :return: The attribute of the network.
        :rtype: Any
        """
        try:
            return super().__getattribute__(name)
        except AttributeError:
            return getattr(self.agent, name)

    def __setattr__(self, name: str, value: Union[Any, AgentType]) -> None:
        if name == "agent" or not hasattr(self, "agent"):
            super().__setattr__(name, value)
        if hasattr(self.agent, name):
            object.__setattr__(self.agent, name, value)
        else:
            super().__setattr__(name, value)

    def clone(self, index: Optional[int] = None, wrap: bool = True) -> SelfAgentWrapper:
        """Clones the wrapper with the underlying agent.

        :param index: Index of the agent in a population, defaults to None
        :type index: Optional[int], optional
        :param wrap: If True, wrap the models in the clone with the accelerator, defaults to False
        :type wrap: bool, optional

        :return: Cloned agent wrapper
        :rtype: SelfAgentWrapper
        """
        agent_clone = self.agent.clone(index, wrap)

        input_args = EvolvableAlgorithm.inspect_attributes(self, input_args_only=True)
        input_args.pop("agent", None)

        clone = self.__class__(agent_clone, **input_args)
        clone = EvolvableAlgorithm.copy_attributes(self, clone)
        return clone

    def save_checkpoint(self, path: str) -> None:
        """Saves a checkpoint of agent properties and network weights to path.

        :param path: Location to save checkpoint at
        :type path: string
        """
        checkpoint = get_checkpoint_dict(self.agent)

        del checkpoint["learn"]
        del checkpoint["get_action"]

        # Add wrapper attributes to checkpoint
        checkpoint["wrapper_cls"] = self.__class__
        checkpoint["wrapper_init_dict"] = EvolvableAlgorithm.inspect_attributes(
            self, input_args_only=True
        )
        checkpoint["wrapper_attrs"] = EvolvableAlgorithm.inspect_attributes(self)

        checkpoint["wrapper_init_dict"].pop("agent")
        checkpoint["wrapper_attrs"].pop("agent")

        # Save checkpoint
        torch.save(
            checkpoint,
            path,
            pickle_module=dill,
        )

    def load_checkpoint(self, path: str) -> None:
        """Loads a checkpoint of agent properties and network weights from path.

        :param path: Location to load checkpoint from
        :type path: string
        """
        checkpoint = torch.load(path, pickle_module=dill, weights_only=False)

        # Load agent properties and network weights
        self.agent.load_checkpoint(path)

        # Load wrapper attributes
        for key, value in checkpoint["wrapper_attrs"].items():
            setattr(self, key, value)

    def get_action(
        self,
        obs: Union[ObservationType, MARLObservationType],
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        """Returns the action from the agent.

        :param obs: Observation from the environment
        :type obs: Union[ObservationType, MARLObservationType]
        :param args: Additional positional arguments
        :type args: Any
        :param kwargs: Additional keyword arguments
        :type kwargs: Any

        :return: Action from the agent
        :rtype: Any
        """
        return self.wrapped_get_action(obs, *args, **kwargs)

    def learn(self, experiences: ExperiencesType, *args: Any, **kwargs: Any) -> Any:
        """Learns from the experiences.

        :param experiences: Experiences from the environment
        :type experiences: ExperiencesType
        :param args: Additional positional arguments
        :type args: Any
        :param kwargs: Additional keyword arguments
        :type kwargs: Any

        :return: Learning information
        :rtype: Any
        """
        return self.wrapped_learn(experiences, *args, **kwargs)


RunningStatsType = Union[
    RunningMeanStd, dict[str, RunningMeanStd], tuple[RunningMeanStd, ...]
]


class RSNorm(AgentWrapper[AgentType]):
    """Wrapper to normalize observations such that each coordinate is centered with unit variance.
    Handles both single and multi-agent settings, as well as Dict and Tuple observation spaces.

    The normalization statistics are only updated when the agent is in training mode. This can be
    disabled during inference through ``agent.set_training_mode(False)``.

    .. warning::
        This wrapper is currently only supported for off-policy algorithms since it relies on
        passed experiences to be formatted as a tuple of PyTorch tensors. Currently
        AgileRL does not use a Buffer class to store experiences for on-policy algorithms, albeit this
        will be released in a soon-to-come update!

    :param agent: Agent to be wrapped
    :type agent: RLAlgorithm, MultiAgentRLAlgorithm
    :param epsilon: Small value to avoid division by zero, defaults to 1e-4
    :type epsilon: float, optional
    :param norm_obs_keys: List of observation keys to normalize, defaults to None
    :type norm_obs_keys: Optional[List]
    """

    obs_rms: Union[RunningStatsType, dict[str, RunningStatsType]]

    def __init__(
        self,
        agent: AgentType,
        epsilon: float = 1e-4,
        norm_obs_keys: Optional[list[str]] = None,
    ) -> None:
        super().__init__(agent)

        self.norm_obs_keys = norm_obs_keys
        if self.multi_agent:
            self.obs_rms = OrderedDict()
            for agent_id, obs_space in self.observation_space.items():
                self.obs_rms[agent_id] = RSNorm.build_rms(
                    obs_space, epsilon, norm_obs_keys, self.device
                )
        else:
            self.obs_rms = RSNorm.build_rms(
                self.observation_space, epsilon, norm_obs_keys, self.device
            )

    @staticmethod
    def build_rms(
        observation_space: spaces.Space,
        epsilon: float = 1e-4,
        norm_obs_keys: Optional[list[str]] = None,
        device: DeviceType = "cpu",
    ) -> Union[RunningMeanStd, dict[str, RunningMeanStd], tuple[RunningMeanStd, ...]]:
        """Builds the RunningMeanStd object(s) based on the observation space.

        :param observation_space: Observation space of the agent
        :type observation_space: spaces.Space
        :return: RunningMeanStd object(s)
        :rtype: Union[RunningMeanStd, dict[str, RunningMeanStd], tuple[RunningMeanStd, ...]]
        """
        if isinstance(observation_space, spaces.Dict):
            if norm_obs_keys is not None:
                observation_space = {
                    key: value
                    for key, value in observation_space.spaces.items()
                    if key in norm_obs_keys
                }

            return {
                key: RunningMeanStd(epsilon, shape=value.shape, device=device)
                for key, value in observation_space.spaces.items()
            }

        elif isinstance(observation_space, spaces.Tuple):
            return tuple(
                RunningMeanStd(epsilon, shape=value.shape, device=device)
                for value in observation_space.spaces
            )

        else:
            return RunningMeanStd(epsilon, shape=observation_space.shape, device=device)

    def _normalize_observation(self, observation: ObservationType) -> ObservationType:
        """Normalizes the observation using the RunningMeanStd object(s).

        :param observation: Observation from the environment
        :type observation: ObservationType

        :return: Normalized observation
        :rtype: ObservationType
        """
        if isinstance(self.obs_rms, dict):
            norm_observation = {}
            for key, rms in self.obs_rms.items():
                norm_observation[key] = (observation[key] - rms.mean) / (
                    rms.var + rms.epsilon
                ).sqrt()

            observation = norm_observation
        elif isinstance(self.obs_rms, tuple):
            norm_observation = []
            for i, rms in enumerate(self.obs_rms):
                norm_obs = (observation[i] - rms.mean) / (rms.var + rms.epsilon).sqrt()
                norm_observation.append(norm_obs)

            observation = tuple(norm_observation)
        else:
            observation = (observation - self.obs_rms.mean) / (
                self.obs_rms.var + self.obs_rms.epsilon
            ).sqrt()

        return observation

    def normalize_observation(self, observation: ObservationType) -> ObservationType:
        """Normalizes the observation using the RunningMeanStd object(s).

        :param observation: Observation from the environment
        :type observation: ObservationType

        :return: Normalized observation
        :rtype: ObservationType
        """
        if self.multi_agent:
            for agent_id, obs in observation.items():
                observation[agent_id] = self._normalize_observation(obs)
            return observation

        return self._normalize_observation(observation)

    def _update_statistics(self, observation: ObservationType) -> None:
        """Updates the running statistics using the observation.

        :param observation: Observation from the environment
        :type observation: ObservationType
        """
        if isinstance(self.obs_rms, dict):
            for key, rms in self.obs_rms.items():
                rms.update(observation[key])
        elif isinstance(self.obs_rms, tuple):
            for i, rms in enumerate(self.obs_rms):
                rms.update(observation[i])
        else:
            self.obs_rms.update(observation)

    def update_statistics(self, observation: ObservationType) -> None:
        """Updates the running statistics using the observation.

        :param observation: Observation from the environment
        :type observation: ObservationType
        """
        if self.multi_agent:
            for _, obs in observation.items():
                self._update_statistics(obs)
        else:
            self._update_statistics(observation)

    def get_action(self, obs: ObservationType, *args: Any, **kwargs: Any) -> Any:
        """Returns the action from the agent after normalizing the observation.

        :param obs: Observation from the environment
        :type obs: ObservationType

        :return: Action from the agent
        :rtype: Any
        """
        obs = obs_to_tensor(obs, self.device)

        # Update running statistics only when in training mode
        if self.training:
            self.update_statistics(obs)

        obs = self.normalize_observation(obs)
        return self.wrapped_get_action(obs, *args, **kwargs)

    def learn(
        self, experiences: Optional[ExperiencesType] = None, *args: Any, **kwargs: Any
    ) -> Any:
        """Learns from the experiences after normalizing the observations.

        :param experiences: Experiences from the environment
        :type experiences: ExperiencesType
        :param args: Additional positional arguments
        :type args: Any
        :param kwargs: Additional keyword arguments
        :type kwargs: Any

        :return: Learning information
        :rtype: Any
        """
        if experiences is None:
            if not isinstance(self.agent, PPO) or not self.agent.use_rollout_buffer:
                raise ValueError(
                    "Experiences must be provided if not using a rollout buffer."
                )

            buffer_size = (
                self.agent.rollout_buffer.capacity
                if not self.agent.rollout_buffer.full
                else self.agent.rollout_buffer.pos
            )
            valid_data = self.agent.rollout_buffer.buffer[:buffer_size]
            valid_data["observations"] = self.normalize_observation(
                valid_data["observations"]
            )
            valid_data["next_observations"] = self.normalize_observation(
                valid_data["next_observations"]
            )
            self.agent.rollout_buffer.buffer[:buffer_size] = valid_data

        # NOTE: We want to move towards always passing experiences as TensorDict objects
        elif is_tensor_collection(experiences):
            experiences["obs"] = self.normalize_observation(experiences["obs"])
            experiences["next_obs"] = self.normalize_observation(
                experiences["next_obs"]
            )
        else:
            experiences = (
                self.normalize_observation(experiences[0]),  # State
                experiences[1],
                experiences[2],
                self.normalize_observation(experiences[3]),  # Next state
                *experiences[4:],
            )

        return self.wrapped_learn(experiences, *args, **kwargs)


class AsyncAgentsWrapper(AgentWrapper[MultiAgentRLAlgorithm]):
    """Wrapper for multi-agent algorithms that solve environments with asynchronous agents (i.e. environments
    where agents don't return observations with the same frequency).

    .. warning::
        This is currently only supported for on-policy multi-agent algorithms such as IPPO.

    :param agent: MultiAgentRLAlgorithm instance to be wrapped.
    :type agent: MultiAgentRLAlgorithm
    """

    def __init__(self, agent: MultiAgentRLAlgorithm) -> None:
        super().__init__(agent)

        assert (
            self.agent.algo == "IPPO"
        ), "AsyncAgentsWrapper is currently only supported for IPPO."

    def extract_inactive_agents(
        self, obs: dict[str, ObservationType]
    ) -> tuple[dict[str, np.ndarray], dict[str, ObservationType]]:
        """Extract the inactive agents from an observation. Inspects each key in the
        observation dictionary and, if all the values are `np.nan` (as set by
        ``AsyncPettingZooVecEnv``), the agent is considered inactive and removed from
        the observation dictionary.

        :param obs: Observation dictionary
        :type obs: dict[str, ObservationType]

        :return: Tuple of inactive agents and filtered observations
        :rtype: tuple[dict[str, np.ndarray], dict[str, ObservationType]]
        """
        inactive_agents = {}
        agents_to_remove = []

        # Process each agent's observations
        for agent_id, agent_obs in obs.items():
            # Get the sample observation based on type
            if isinstance(agent_obs, dict):
                sample = next(iter(agent_obs.values()))
            elif isinstance(agent_obs, tuple):
                sample = agent_obs[0]
            else:
                sample = agent_obs

            # Skip non-vectorized environments, assuming env doesn't return
            # observations for inactive agents
            if len(sample.shape) == 1:
                continue

            # Create boolean mask for active agents
            active_mask: np.ndarray = ~np.isnan(sample).all(axis=1)

            # If all agents are active, skip
            if active_mask.all():
                continue

            # Get indices of inactive agents
            inactive_agent_indices = np.where(~active_mask)[0]

            # If all agents are inactive, mark for removal
            if not active_mask.any():
                agents_to_remove.append(agent_id)
                continue

            # Apply mask to filter observations
            if isinstance(agent_obs, dict):
                obs[agent_id] = {k: v[active_mask] for k, v in agent_obs.items()}
            elif isinstance(agent_obs, tuple):
                obs[agent_id] = tuple(v[active_mask] for v in agent_obs)
            else:
                obs[agent_id] = agent_obs[active_mask]

            inactive_agents[agent_id] = inactive_agent_indices

        # Remove completely inactive agents
        for agent_id in agents_to_remove:
            obs.pop(agent_id)

        return inactive_agents, obs

    def stack_experiences(self, experience: ExperiencesType) -> ExperiencesType:
        """Stacks the experiences.

        :param experiences: Experiences from the environment
        :type experiences: ExperiencesType
        """
        stacked_experience = {}
        for agent_id, inp in experience.items():
            if isinstance(inp, list):
                stacked_exp = (
                    stack_experiences(inp, to_torch=False)[0] if len(inp) > 0 else None
                )
            else:
                stacked_exp = inp

            stacked_experience[agent_id] = stacked_exp

        return stacked_experience

    def get_action(
        self, obs: ObservationType, *args: Any, **kwargs: Any
    ) -> ActionReturnType:
        """Returns the action from the agent. Since the environments may not return observations for all agents
        at the same time, we need to extract the inactive agents from the observation and fill in placeholder
        values for their actions.

        :param obs: Observation from the environment
        :type obs: ObservationType

        :return: Action from the agent
        :rtype: Any
        """
        inactive_agents, obs = self.extract_inactive_agents(obs)
        action_return: ActionReturnType = self.wrapped_get_action(obs, *args, **kwargs)

        # Need to fill in placeholder np.nan for inactive agents
        action_dict = (
            action_return[0] if isinstance(action_return, tuple) else action_return
        )
        for agent_id, inactive_array in inactive_agents.items():
            placeholder = (
                int(np.nan)
                if np.issubdtype(action_dict[agent_id].dtype, np.integer)
                else np.nan
            )

            # Insert placeholder values for inactive agents
            action_dict[agent_id] = np.insert(
                action_dict[agent_id], inactive_array, placeholder, axis=0
            )

        if isinstance(action_return, tuple):
            action_return = (action_dict,) + action_return[1:]
        else:
            action_return = action_dict

        return action_return

    def learn(self, experiences: ExperiencesType, *args: Any, **kwargs: Any) -> Any:
        """Learns from the collected experiences.

        :param experiences: Experiences from the environment
        :type experiences: ExperiencesType
        :param args: Additional positional arguments
        :type args: Any
        :param kwargs: Additional keyword arguments
        :type kwargs: Any

        :return: Learning information
        :rtype: Any
        """
        (states, actions, log_probs, rewards, dones, values, next_state, next_done) = (
            map(self.stack_experiences, experiences)
        )

        # Handle case where we haven't collected a next state for each sub-agent
        for agent_id in self.agent.agent_ids:
            agent_next_state: Optional[np.ndarray] = next_state.get(agent_id, None)

            # If we haven't collected a next state for this agent yet, we need to use
            # last collected state as next_state
            if agent_next_state is None or np.isnan(agent_next_state).all():
                agent_states = states[agent_id]
                agent_dones = dones[agent_id]
                agent_rewards = rewards[agent_id]

                # Update to use last collected state as next_state
                next_state[agent_id] = agent_states[-1]
                next_done[agent_id] = agent_dones[-1]
                states[agent_id] = agent_states[:-1]
                dones[agent_id] = agent_dones[:-1]
                rewards[agent_id] = agent_rewards[:-1]
                actions[agent_id] = actions[agent_id][:-1]
                log_probs[agent_id] = log_probs[agent_id][:-1]
                values[agent_id] = values[agent_id][:-1]

        experiences = (
            states,
            actions,
            log_probs,
            rewards,
            dones,
            values,
            next_state,
            next_done,
        )
        return self.wrapped_learn(experiences, *args, **kwargs)



import os

import accelerate

from agilerl.utils.ilql_utils import (
    add_system_configs,
    convert_path,
    strip_from_beginning,
    strip_from_end,
    to_bin,
)


def test_convert_path_none():
    path = None
    converted_path = convert_path(path)
    assert converted_path is None


def test_convert_path():
    path = "example_path"
    converted_path = convert_path(path)
    assert converted_path == os.path.join(
        os.path.dirname(os.path.realpath("agilerl/utils/ilql_utils.py")), "../../", path
    )


def test_add_system_configs():
    config = {}
    accelerator = accelerate.Accelerator()
    config = add_system_configs(config, accelerator)

    assert config["device"] == str(accelerator.device)
    assert config["num_processes"] == accelerator.num_processes
    assert config["use_fp16"] == (accelerator.mixed_precision != "no")


def test_to_bin_none():
    n = 10
    pad_to_size = None

    bins = to_bin(n, pad_to_size)

    assert bins == [1, 0, 1, 0]


def test_to_bin():
    n = 10
    pad_to_size = 5

    bins = to_bin(n, pad_to_size)

    assert bins == [0, 1, 0, 1, 0]


def test_strip_from_end():
    string = "this string will be stripped from the end"
    strip_key = "from the end"

    string = strip_from_end(string, strip_key)

    assert string == "this string will be stripped "


def test_strip_from_beginning():
    string = "this string will be stripped from the beginning"
    strip_key = "this string will be stripped "

    string = strip_from_beginning(string, strip_key)

    assert string == "from the beginning"


def test_no_strip_from_beginning():
    string = "this string will not be stripped"
    strip_key = "zzz"

    string = strip_from_beginning(string, strip_key)

    assert string == "this string will not be stripped"



import copy
from collections import OrderedDict
from dataclasses import asdict
from typing import Any, Optional, TypeVar, Union

import torch
import torch.nn as nn
from gymnasium import spaces

from agilerl.modules import EvolvableCNN, EvolvableLSTM, EvolvableMLP
from agilerl.modules.base import EvolvableModule, ModuleDict, MutationType, mutation
from agilerl.modules.configs import CnnNetConfig, MlpNetConfig, NetConfig
from agilerl.typing import ArrayOrTensor, ConfigType, ModuleType, NetConfigType
from agilerl.utils.evolvable_networks import (
    get_activation,
    is_image_space,
    is_vector_space,
    tuple_to_dict_space,
)

SelfMultiInput = TypeVar("SelfMultiInput", bound="EvolvableMultiInput")
SupportedEncoderTypes = Union[EvolvableCNN, EvolvableMLP, EvolvableLSTM, SelfMultiInput]
MultiInputConfigType = Union[ConfigType, dict[str, ConfigType]]
TupleOrDictSpace = Union[spaces.Tuple, spaces.Dict]
TupleOrDictObservation = Union[dict[str, ArrayOrTensor], tuple[ArrayOrTensor]]

# Default configurations for the feature extractors
DefaultCnnConfig = CnnNetConfig(
    channel_size=[16, 16],
    kernel_size=[3, 3],
    stride_size=[1, 1],
    output_activation="ReLU",
)
DefaultMlpConfig = MlpNetConfig(
    hidden_size=[64, 64],
    output_activation="ReLU",
    output_layernorm=True,
)


def get_total_flatdim(observation_spaces: spaces.Dict) -> int:
    """Get the total flat dimension of the observation space.

    :param observation_space: Dictionary or Tuple space of observations.
    :type observation_space: spaces.Dict or spaces.Tuple
    :return: Total flat dimension of the observation space.
    """
    return sum([spaces.flatdim(space) for space in observation_spaces.spaces.values()])


def vector_space_check(space: spaces.Space) -> bool:
    """Check if the space is a vector space as it pertains to the use of
    an EvolvableMLP.

    :param space: Input space
    :type space: spaces.Space
    :return: True if the space is a vector space, False otherwise
    """
    return is_vector_space(space) or (
        isinstance(space, spaces.Box) and len(space.shape) > 3
    )


class EvolvableMultiInput(EvolvableModule):
    """Evolvable multi-input network for Tuple or Dict observation spaces. It inspects the
    observation space to determine the type of network to build for each key. It builds an
    ``EvolvableCNN`` for image subspaces and a nn.Flatten() for other types. Vector observations are
    concatenated with the extracted features before passing through an ``EvolvableMLP`` to produce
    the output tensor. Optionally, users may specify an additional ``EvolvableMLP`` to be applied to
    the concatenated vector observations before concatenation with the extracted features.

    Supports the following types of architecture mutations during training:

    * Adding or removing latent nodes
    * Inherits the mutation methods of any nested ``EvolvableModule`` objects used in the network

    :param observation_space: Dictionary or Tuple space of observations.
    :type observation_space: spaces.Dict or spaces.Tuple
    :param num_outputs: Dimension of the output tensor.
    :type num_outputs: int
    :param latent_dim: Dimension of the latent space representation. Default is 16.
    :type latent_dim: int, optional
    :param vector_space_mlp: Whether to use an MLP for the vector spaces. This is done by concatenating the
        flattened observations and passing them through an `EvolvableMLP`. Default is False, whereby the
        observations are concatenated directly to the feature encodings before the final MLP.
    :type vector_space_mlp: bool, optional
    :param cnn_config: Configuration for the CNN feature extractor. Default is None.
    :type cnn_config: MultiInputConfigType, optional
    :param mlp_config: Configuration for the MLP feature extractor. Default is None.
    :type mlp_config: MultiInputConfigType, optional
    :param init_dicts: Dictionary of initialization dictionaries for the feature extractors. Default is {}.
    :type init_dicts: dict[str, dict[str, Any]], optional
    :param output_activation: Activation function for the output layer. Default is None.
    :type output_activation: Optional[str], optional
    :param min_latent_dim: Minimum dimension of the latent space. Default is 8.
    :type min_latent_dim: int, optional
    :param max_latent_dim: Maximum dimension of the latent space. Default is 128.
    :type max_latent_dim: int, optional
    :param device: Device to use for the network. Default is "cpu".
    :type device: str, optional
    :param name: Name of the network. Default is "multi_input".
    :type name: str, optional
    :param random_seed: Random seed to use for the network. Defaults to None.
    :type random_seed: Optional[int]
    """

    feature_net: ModuleDict

    def __init__(
        self,
        observation_space: TupleOrDictSpace,
        num_outputs: int,
        latent_dim: int = 32,
        vector_space_mlp: bool = False,
        cnn_config: Optional[MultiInputConfigType] = None,
        mlp_config: Optional[MultiInputConfigType] = None,
        init_dicts: Optional[MultiInputConfigType] = None,
        output_activation: Optional[str] = None,
        output_layernorm: bool = False,
        min_latent_dim: int = 8,
        max_latent_dim: int = 128,
        device: str = "cpu",
        name: str = "multi_input",
        random_seed: Optional[int] = None,
    ):
        super().__init__(device, random_seed)

        assert num_outputs > 0, "Number of outputs must be greater than 0."
        assert latent_dim > 0, "Latent dimension must be greater than 0."
        assert isinstance(
            observation_space, (spaces.Dict, spaces.Tuple)
        ), "Observation space must be a Dict or Tuple space."
        assert (
            latent_dim <= max_latent_dim
        ), "Latent dimension must be less than or equal to max latent dimension."
        assert (
            latent_dim >= min_latent_dim
        ), "Latent dimension must be greater than or equal to min latent dimension."

        # Convert Tuple space to Dict space for consistency
        self.is_tuple_space = False
        if isinstance(observation_space, spaces.Tuple):
            observation_space = tuple_to_dict_space(observation_space)
            self.is_tuple_space = True

        self.observation_space = observation_space
        self.num_outputs = num_outputs
        self.cnn_config = cnn_config or DefaultCnnConfig
        self.mlp_config = mlp_config or DefaultMlpConfig
        self._init_dicts = init_dicts or {}
        self._activation = None
        self.mlp_name = None
        self.vector_space_mlp = vector_space_mlp
        self.latent_dim = latent_dim
        self.output_activation = output_activation
        self.min_latent_dim = min_latent_dim
        self.max_latent_dim = max_latent_dim
        self.name = name
        self.device = device

        # We use an output nn.LayerNorm whenever specified, or if there is an MLP
        # feature extractor that uses layer_norm=True
        self.output_layernorm = output_layernorm or (
            self.vector_space_mlp and self.mlp_config.get("layer_norm", True)
        )

        # Modifications for MLP encoders
        if self.vector_space_mlp:
            self._modify_mlp_config()

        # Vector spaces
        self.vector_spaces = spaces.Dict(
            {
                key: space
                for key, space in observation_space.spaces.items()
                if vector_space_check(space)
            }
        )
        self.total_vector_dims = get_total_flatdim(self.vector_spaces)

        # Build feature extractor
        self.feature_net = self.build_feature_extractor()

        # Calculate total extracted features from non-vector spaces
        # (i.e. Box spaces with more than one dimension)
        self.extracted_features_dim = self.calc_extracted_features_dim()

        # Vector observations (i.e. 1D Box, Discrete, MultiDiscrete) are either
        # passed through an MLP or concatenated directly to the extracted features
        features_dim = self.extracted_features_dim + self.total_vector_dims * (
            1 - self.vector_space_mlp
        )
        # Final dense layer to convert feature encodings to desired num_outputs
        self.final_dense = nn.Linear(features_dim, num_outputs, device=device)
        self.final_layernorm = (
            nn.LayerNorm(num_outputs, device=device, elementwise_affine=False)
            if self.output_layernorm
            else None
        )
        self.output = get_activation(output_activation)

    @property
    def net_config(self) -> dict[str, Any]:
        """Returns the configuration of the network.

        :return: Network configuration
        :rtype: dict[str, Any]
        """
        net_config = self.init_dict.copy()
        for attr in ["observation_space", "num_outputs", "device", "name"]:
            net_config.pop(attr)
        return net_config

    @property
    def activation(self) -> str:
        """Get the activation function for the network.

        :return: Activation function
        :rtype: str
        """
        return self._activation

    @activation.setter
    def activation(self, activation: str) -> None:
        """Set the activation function for the network.

        :param activation: Activation function to use.
        :type activation: str
        """
        self._activation = activation
        self.feature_net.change_activation(activation, output=True)

    @property
    def init_dicts(self) -> dict[str, dict[str, Any]]:
        """Returns the initialization dictionaries for the network.

        :return: Initialization dictionaries
        :rtype: dict[str, dict[str, Any]]
        """
        if not hasattr(self, "feature_net"):
            return self._init_dicts

        reformatted_dicts = {}
        for key, net in self.feature_net.modules().items():
            init_dict = net.init_dict
            init_dict.pop("observation_space", None)  # MultiInput
            init_dict.pop("input_size", None)  # LSTM
            init_dict.pop("input_shape", None)  # CNN
            init_dict.pop("num_inputs", None)  # MLP
            init_dict["num_outputs"] = self.latent_dim
            reformatted_dicts[key] = init_dict

        return reformatted_dicts

    @property
    def cnn_init_dict(self) -> dict[str, Any]:
        """Returns the initialization dictionary for the CNN."""
        return copy.deepcopy(self.cnn_config)

    @property
    def mlp_init_dict(self) -> dict[str, Any]:
        """Returns the initialization dictionary for the MLP."""
        return copy.deepcopy(self.mlp_config)

    def _reformat_mlp_config(self, config: dict[str, Any]) -> dict[str, Any]:
        """Reformats the MLP configuration."""
        config["output_vanish"] = False  # Don't want output vanish for encoder
        config["output_layernorm"] = config.get(
            "layer_norm", True
        )  # Add layer norm for output if present generally
        config["output_activation"] = config.get(
            "activation", "ReLU"
        )  # Use same output activation
        return config

    def _modify_mlp_config(self) -> None:
        """Modifies the MLP architecture to be appropriate for use as an encoder (i.e. disable
        output vanishing, and apply layer normalization at the final layer consistently with the
        rest of the network). See https://github.com/AgileRL/AgileRL/issues/337 for more details.
        """
        self.mlp_config["output_vanish"] = False
        self.mlp_config["output_layernorm"] = self.mlp_config.get("layer_norm", True)
        self.mlp_config["output_activation"] = self.mlp_config.get("activation", "ReLU")

    def init_weights_gaussian(
        self, std_coeff: float = 4, output_coeff: float = 4
    ) -> None:
        """Initialise weights of linear layers using Gaussian distribution."""
        for module in self.feature_net.modules().values():
            module.init_weights_gaussian(std_coeff=std_coeff)

        # Initialise final dense layer
        EvolvableModule.init_weights_gaussian(self.final_dense, std_coeff=output_coeff)

    def calc_extracted_features_dim(self) -> int:
        """Calculates the toal dimension of the features extracted by the evolvable
        feature extractors.

        :return: Total dimension of the extracted features.
        :rtype: int
        """
        return sum(
            [
                self.latent_dim
                for name in self.feature_net.keys()
                if name not in self.vector_spaces.keys()
            ]
        )

    def get_inner_init_dict(self, key: str, default: ModuleType) -> NetConfigType:
        """Returns the initialization dictionary for the specified key.

        :param key: Key of the observation space.
        :type key: str
        :param default: Default value to return if the key is not found.
        :type default: ModuleType
        :return: Initialization dictionary.
        :rtype: ConfigType
        """
        init_dicts = self.init_dicts
        if key in init_dicts:
            init_dict = init_dicts[key]
            init_dict["num_outputs"] = self.latent_dim
            init_dict["device"] = self.device
            return init_dict

        init_dict = {
            ModuleType.CNN: self.cnn_init_dict,
            ModuleType.MLP: self.mlp_init_dict,
            ModuleType.MULTI_INPUT: self.net_config,
        }.get(default)

        if init_dict is None:
            raise ValueError(
                "Invalid default value provided, must be 'cnn' or 'mlp' or 'multi_input'."
            )
        else:
            # Check if we are extracting a nested dict
            nested_dict = init_dict.get(key)
            init_dict = (
                copy.deepcopy(nested_dict)
                if nested_dict is not None
                else copy.deepcopy(init_dict)
            )

            if isinstance(init_dict, NetConfig):
                init_dict = asdict(init_dict)

        init_dict["num_outputs"] = self.latent_dim
        init_dict["device"] = self.device
        return init_dict

    def build_feature_extractor(self) -> dict[str, SupportedEncoderTypes]:
        """Creates the feature extractor and final MLP networks.

        :return: Dictionary of feature extractors.
        :rtype: dict[str, EvolvableMLP | EvolvableCNN | EvolvableLSTM | EvolvableMultiInput]
        """
        # Automatically build feature extractors from subspaces
        feature_net = ModuleDict(device=self.device)
        for key, space in self.observation_space.spaces.items():
            if isinstance(space, spaces.Box) and len(space.shape) in [0, 1]:
                continue

            # EvolvableMultiInput for nested multi-input spaces
            if isinstance(space, (spaces.Dict, spaces.Tuple)):
                init_dict = self.get_inner_init_dict(
                    key, default=ModuleType.MULTI_INPUT
                )
                feature_extractor = EvolvableMultiInput(
                    observation_space=space,
                    name=init_dict.pop("name", key),
                    **init_dict,
                )

            # EvolvableCNN for image spaces
            elif is_image_space(space):
                init_dict = self.get_inner_init_dict(key, default=ModuleType.CNN)
                feature_extractor = EvolvableCNN(
                    input_shape=space.shape,
                    name=init_dict.pop("name", key),
                    **init_dict,
                )
            else:  # Flatten other observation types
                feature_extractor = nn.Flatten()

            feature_net[key] = feature_extractor

        # Optionally, use an EvolvableMLP for all concatenated vector inputs
        if self.vector_space_mlp:
            init_dict = self.get_inner_init_dict("vector_mlp", default=ModuleType.MLP)

            self.mlp_name = init_dict.pop("name", "vector_mlp")
            vector_mlp = EvolvableMLP(
                num_inputs=self.total_vector_dims, name=self.mlp_name, **init_dict
            )
            feature_net[self.mlp_name] = vector_mlp

        return feature_net

    def forward(self, x: TupleOrDictObservation) -> torch.Tensor:
        """Forward pass of the composed network. Extracts features from each observation key and concatenates
        them with the corresponding observation key if specified. The concatenated features are then passed
        through the final MLP to produce the output tensor.

        :param x: Dictionary of observations.
        :type x: dict[str, ArrayOrTensor], tuple[ArrayOrTensor]
        :return: Output tensor.
        :rtype: torch.Tensor
        """
        if isinstance(x, tuple):
            x = dict(zip(self.observation_space.spaces.keys(), x))

        for key, obs in x.items():
            if not isinstance(obs, torch.Tensor):
                x[key] = torch.tensor(obs, device=self.device, dtype=torch.float32)

        # Extract features from non-vector subspaces
        extracted_features = OrderedDict()
        if self.extracted_features_dim > 0:
            for key in x.keys():
                if key in self.feature_net.keys():
                    extracted_features[key] = self.feature_net[key](x[key])

        # Extract raw features from vector spaces
        vector_inputs = []
        for key, space in self.vector_spaces.items():
            _obs = (
                extracted_features.pop(key)
                if key in extracted_features.keys()
                else x[key]
            )
            if len(_obs.shape) == 1:
                dim = len(space.shape) - 1
                _obs = _obs.unsqueeze(dim)

            vector_inputs.append(_obs)

        # Concatenate vector inputs and, optionally, pass through additional EvolvableMLP
        vector_inputs = (
            torch.cat(vector_inputs, dim=1)
            if vector_inputs
            else torch.tensor([], device=self.device)
        )
        vector_features = (
            self.feature_net[self.mlp_name](vector_inputs)
            if self.vector_space_mlp
            else vector_inputs
        )

        # Concatenate extracted latent features
        extracted_features = (
            torch.cat(list(extracted_features.values()), dim=1)
            if extracted_features
            else torch.tensor([], device=self.device)
        )

        # Concatenate all features and pass through final MLP
        features = torch.cat([extracted_features, vector_features], dim=1)
        latent = self.final_dense(features)

        if self.output_layernorm:
            latent = self.final_layernorm(latent)

        return self.output(latent)

    @mutation(MutationType.ACTIVATION)
    def change_activation(self, activation: str, output: bool = False) -> None:
        """Set the activation function for the network.

        :param activation: Activation function to use.
        :type activation: str
        :param output: Flag indicating whether to set the output activation function, defaults to False
        :type output: bool, optional

        :return: Activation function
        :rtype: str
        """
        self.activation = activation
        if output:
            self.output = get_activation(activation)

    @mutation(MutationType.NODE)
    def add_latent_node(self, numb_new_nodes: Optional[int] = None) -> dict[str, Any]:
        """Add a latent node to the network.

        :param numb_new_nodes: Number of new nodes to add, defaults to None
        :type numb_new_nodes: int, optional

        :return: Dictionary specifying the number of nodes added.
        :rtype: dict[str, Any]
        """
        if numb_new_nodes is None:
            numb_new_nodes = self.rng.choice([8, 16, 32])

        if self.latent_dim + numb_new_nodes < self.max_latent_dim:
            self.latent_dim += numb_new_nodes

        return {"numb_new_nodes": numb_new_nodes}

    @mutation(MutationType.NODE)
    def remove_latent_node(
        self, numb_new_nodes: Optional[int] = None
    ) -> dict[str, Any]:
        """Remove a latent node from the network.

        :param numb_new_nodes: Number of nodes to remove, defaults to None
        :type numb_new_nodes: int, optional

        :return: Dictionary specifying the number of nodes removed.
        :rtype: dict[str, Any]
        """
        if numb_new_nodes is None:
            numb_new_nodes = self.rng.choice([8, 16, 32])

        if self.latent_dim - numb_new_nodes > self.min_latent_dim:
            self.latent_dim -= numb_new_nodes

        return {"numb_new_nodes": numb_new_nodes}

    def recreate_network(self) -> None:
        """Recreates the network with the new latent dimension."""
        feature_net = self.build_feature_extractor()
        self.feature_net = EvolvableModule.preserve_parameters(
            old_net=self.feature_net, new_net=feature_net
        )

        # Calculate total extracted features dimension
        extracted_features_dim = self.calc_extracted_features_dim()
        features_dim = extracted_features_dim + self.total_vector_dims * (
            1 - self.vector_space_mlp
        )
        final_dense = nn.Linear(features_dim, self.num_outputs, device=self.device)
        self.final_dense = EvolvableModule.preserve_parameters(
            old_net=self.final_dense, new_net=final_dense
        )



# AgileRL On-policy (RNN/MLP) MiniGrid Memory Demo
#
# This script demonstrates how to use recurrent neural networks (RNNs) or MLPs with PPO to solve the MiniGrid-DoorKey-8x8-v0 environment.
# The observation wrapper flattens the image and concatenates a one-hot encoding of the agent's direction.
# This version follows the training structure of performance_flamegraph_cartpole.py and performance_flamegraph_lunar_lander.py,
# using a population and a simple evolutionary loop.
import os
import shutil
from typing import Any, Callable, Optional

import gymnasium as gym
import imageio

# import minigrid NOTE: Remove comment when training
import numpy as np
import torch

from agilerl.algorithms import PPO
from agilerl.hpo.mutation import Mutations
from agilerl.hpo.tournament import TournamentSelection
from agilerl.rollouts.on_policy import collect_rollouts, collect_rollouts_recurrent
from agilerl.typing import BPTTSequenceType
from agilerl.utils.utils import create_population, default_progress_bar


# --- Define the MiniGrid Observation Wrapper ---
class MiniGridObsWrapper(gym.ObservationWrapper):
    """
    Extracts and flattens the 'image' observation from MiniGrid, normalizes to [0, 1],
    and concatenates a one-hot encoding of the 'direction' field.
    """

    def __init__(self, env: gym.Env):
        super().__init__(env)
        img_shape = self.observation_space["image"].shape
        flat_img_dim = np.prod(img_shape)
        # The new observation space is flattened image + 4 for direction one-hot
        self.observation_space = gym.spaces.Box(
            low=0.0, high=1.0, shape=(flat_img_dim + 4,), dtype=np.float32
        )

    def reset(self, **kwargs: Any) -> tuple[np.ndarray, dict[str, Any]]:
        obs, info = self.env.reset(**kwargs)
        return self.observation(obs), info

    def observation(self, obs: dict[str, np.ndarray]) -> np.ndarray:
        img = obs["image"].astype(np.float32)
        flat_img = img.flatten()
        direction = obs["direction"]
        direction_onehot = np.zeros(4, dtype=np.float32)
        direction_onehot[direction] = 1.0
        concat_obs = np.concatenate([flat_img, direction_onehot], axis=0)
        return concat_obs


def run_demo():
    # --- Setup Configuration ---
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    print(f"Using device: {device}")

    # Toggle this to True for RNN (LSTM), False for MLP
    recurrent = True  # <--- CHANGE THIS TO ENABLE/DISABLE RECURRENT
    active_collect = collect_rollouts if not recurrent else collect_rollouts_recurrent

    if recurrent:
        NET_CONFIG = {
            "latent_dim": 64,
            "encoder_config": {
                "hidden_state_size": 64,  # LSTM hidden state size
            },
            "head_config": {"hidden_size": [64]},
        }
    else:
        NET_CONFIG = {
            "encoder_config": {
                "hidden_size": [128],
            },
        }

    # Hyperparameters
    num_envs = 16  # Fewer envs for MiniGrid due to slowness
    learn_step = 512 * num_envs
    INIT_HP = {
        "POP_SIZE": 1,  # Population size
        "BATCH_SIZE": 128,
        "LEARN_STEP": learn_step,
        "LR": 1e-4,
        "GAMMA": 0.99,
        "GAE_LAMBDA": 0.95,
        "CLIP_COEF": 0.2,
        "ENT_COEF": 0.01,
        "VF_COEF": 0.5,
        "MAX_GRAD_NORM": 0.5,
        "UPDATE_EPOCHS": 4,
        "HIDDEN_STATE_SIZE": 128,
        "SHARE_ENCODERS": False,
        "USE_ROLLOUT_BUFFER": True,
        "RECURRENT": recurrent,
        "BPTT_SEQUENCE_TYPE": BPTTSequenceType.CHUNKED,
        "MAX_SEQ_LEN": 16,  # Maximum sequence length for truncated BPTT
        "ACTION_STD_INIT": 0.6,
        "TARGET_KL": None,
    }

    # --- Create Environment and Population ---
    def make_env(render_mode: Optional[str] = None) -> Callable[[], gym.Env]:
        env = gym.make("MiniGrid-DoorKey-8x8-v0", render_mode=render_mode)
        env = MiniGridObsWrapper(env)
        return env

    env = gym.vector.SyncVectorEnv([make_env for _ in range(num_envs)])
    single_test_env = gym.vector.SyncVectorEnv([make_env])

    observation_space = env.single_observation_space
    action_space = env.single_action_space

    pop: list[PPO] = create_population(
        algo="PPO",
        observation_space=observation_space,
        action_space=action_space,
        net_config=NET_CONFIG,
        INIT_HP=INIT_HP,
        population_size=INIT_HP["POP_SIZE"],
        num_envs=num_envs,
        device=device,
    )

    # --- Setup Evolution Components ---
    tournament = TournamentSelection(
        tournament_size=2,
        elitism=True,
        population_size=INIT_HP["POP_SIZE"],
        eval_loop=1,
    )

    mutations = Mutations(
        no_mutation=0.4,
        architecture=0,
        new_layer_prob=0.0,
        parameters=0.2,
        activation=0,
        rl_hp=0.2,
        mutation_sd=0.1,
        activation_selection=["ReLU", "ELU", "GELU"],
        mutate_elite=True,
        rand_seed=1,
        device=device,
    )

    # --- Training Loop (Performance-Flamegraph Style) ---
    max_steps = 5_000_000 // len(pop)
    required_score = 0.9
    evo_steps = INIT_HP["LEARN_STEP"] * 5
    eval_steps = None
    eval_loop = 5

    total_steps = 0
    training_complete = False

    print("Training...")
    pbar = default_progress_bar(max_steps)
    while (
        np.less([agent.steps[-1] for agent in pop], max_steps).all()
        and not training_complete
    ):
        pop_episode_scores = []
        for agent in pop:
            steps = 0
            completed_episodes = []
            last_obs, last_done, last_scores, last_info = None, None, None, None
            for _ in range(-(evo_steps // -agent.learn_step)):
                # Collect rollouts and save in buffer
                episode_scores, last_obs, last_done, last_scores, last_info = (
                    active_collect(
                        agent,
                        env,
                        last_obs=last_obs,
                        last_done=last_done,
                        last_scores=last_scores,
                        last_info=last_info,
                    )
                )

                agent.learn()  # Learn from rollout buffer

                # Update step counter and scores
                total_steps += agent.learn_step
                steps += agent.learn_step
                agent.steps[-1] += agent.learn_step
                completed_episodes += episode_scores

                pbar.write(f"Completed scores: {round(np.mean(episode_scores), 2)}")

            pop_episode_scores.append(
                np.mean(completed_episodes)
                if len(completed_episodes) > 0
                else "0 completed episodes"
            )
            pbar.update(steps // len(pop))

        # Evaluate and evolve
        fitnesses = [
            agent.test(
                single_test_env,
                max_steps=eval_steps,
                loop=eval_loop,
            )
            for agent in pop
        ]
        pbar.write(
            f"--- Global steps {total_steps} ---\n"
            f"Steps: {[agent.steps[-1] for agent in pop]}\n"
            f"Scores: {pop_episode_scores}\n"
            f"Fitnesses: {['%.2f' % fitness for fitness in fitnesses]}\n"
            f"5 fitness avgs: {['%.2f' % np.mean(agent.fitness[-5:]) for agent in pop]}\n"
            f"Mutations: {[agent.mut for agent in pop]}\n"
        )

        if any(score >= required_score for score in pop_episode_scores):
            print(
                f"\nAgent achieved required score {required_score}. Stopping training."
            )
            elite, _ = tournament.select(pop)
            training_complete = True
            break

        elite, pop = tournament.select(pop)
        pop = mutations.mutation(pop)
        for agent in pop:
            agent.steps.append(agent.steps[-1])

    pbar.close()
    env.close()

    # --- Record GIF of Best Agent ---
    print("Recording GIF of best agent...")

    gifs_dir = "gifs"
    os.makedirs(gifs_dir, exist_ok=True)
    env_to_wrap = make_env(render_mode="rgb_array")()

    if not training_complete:
        fitnesses = [
            agent.test(
                env_to_wrap,
                max_steps=eval_steps,
                loop=eval_loop,
                vectorized=False,
            )
            for agent in pop
        ]
        elite, _ = tournament.select(pop)

    render_env = gym.wrappers.RecordVideo(
        env_to_wrap, video_folder="temp_video", disable_logger=True
    )

    total_steps = 0
    episode_rewards = []

    for episode in range(3):
        obs, _ = render_env.reset()
        done = False
        episode_reward = 0
        episode_steps = 0
        if recurrent:
            hidden_state = elite.get_initial_hidden_state(1)

        episode_frames = []
        while not done:
            frame = render_env.render()
            episode_frames.append(frame)
            if recurrent:
                action, _, _, _, hidden_state = elite.get_action(
                    obs, hidden_state=hidden_state
                )
            else:
                action, _, _, _, _ = elite.get_action(obs)
            obs, reward, terminated, truncated, _ = render_env.step(action)
            done = terminated or truncated
            episode_reward += reward
            episode_steps += 1

        if episode_frames:
            gif_path = os.path.join(
                gifs_dir,
                f"minigrid_memory_{'rnn' if recurrent else 'mlp'}_episode_{episode + 1}.gif",
            )
            imageio.mimsave(gif_path, episode_frames, fps=15)
            print(f"Saved GIF for episode {episode + 1} to {gif_path}")

        total_steps += episode_steps
        episode_rewards.append(episode_reward)
        print(
            f"Recorded Episode {episode + 1} Reward: {episode_reward}, Steps: {episode_steps}"
        )

    avg_reward = sum(episode_rewards) / len(episode_rewards)
    avg_steps = total_steps / len(episode_rewards)
    print(f"Average Reward: {avg_reward:.2f}, Average Steps: {avg_steps:.2f}")

    render_env.close()

    if os.path.exists("temp_video"):
        shutil.rmtree("temp_video")

    print(f"GIFs saved to {os.path.abspath(gifs_dir)}")


if __name__ == "__main__":
    run_demo()



import gc
from typing import Any

import numpy as np
import torch
import torch.nn.functional as F
from accelerate import Accelerator

from agilerl.algorithms.core.base import LLMAlgorithm
from agilerl.algorithms.core.registry import HyperparameterConfig, NetworkGroup
from agilerl.protocols import LoraConfigProtocol, PreTrainedModelProtocol
from agilerl.typing import ExperiencesType, LLMObsType
from agilerl.utils.algo_utils import get_experiences_samples
from agilerl.utils.llm_utils import PreferenceGym


class DPO(LLMAlgorithm):
    """The DPO algorithm class. DPO paper: https://arxiv.org/pdf/2305.18290

    :param pad_token_id: Pad token id
    :type pad_token_id: int
    :param pad_token: Pad token
    :type pad_token: str
    :param model_name: Model name
    :type model_name: str, optional
    :param actor_network: HuggingFace LLM
    :type actor_network: PreTrainedModelProtocol
    :param model_config: Model configuration, to be used when creating the model from a name or path
    :param hp_config: RL hyperparameter mutation configuration, defaults to None, whereby algorithm mutations are disabled.
    :type hp_config: HyperparameterConfig, optional
    :param index: Index to keep track of object instance during tournament selection and mutation, defaults to 0
    :type index: int, optional
    :param batch_size: Batch size for training, defaults to 16
    :type batch_size: int, optional
    :param lr: Learning rate, defaults to 0.000005
    :type lr: float, optional
    :param beta: Beta parameter for DPO, defaults to 0.001
    :type beta: float, optional
    :param max_grad_norm: Maximum gradient norm, defaults to 0.1
    :type max_grad_norm: float, optional
    :param update_epochs: Number of update epochs, defaults to 1
    :type update_epochs: int, optional
    :param calc_position_embeddings: Flag to indicate if position embeddings should be calculated, defaults to True
    :type calc_position_embeddings: bool, optional
    :param micro_batch_size_per_gpu: Micro batch size per GPU, defaults to None
    :type micro_batch_size_per_gpu: int, optional
    :param reduce_memory_peak: Flag to indicate if memory peak should be reduced, defaults to False
    :type reduce_memory_peak: bool, optional
    :param device: Device for accelerated computing, 'cpu' or 'cuda', defaults to 'cpu'
    :type device: str, optional
    :param lora_config: Config for LoRA, defaults to None
    :type lora_config: LoraConfigProtocol, optional
    :param accelerator: Accelerator for distributed computing, defaults to None
    :type accelerator: accelerate.Accelerator(), optional
    :param wrap: Wrap models for distributed training upon creation, defaults to True
    :type wrap: bool, optional
    :param clone: Flag to indicate if the instantiation is a cloning, defaults to False
    :type clone: bool, optional
    :param use_separate_reference_adapter: Flag to indicate if the reference policy should have a separate adapter, defaults to False
    :type use_separate_reference_adapter: bool, optional
    :param seed: Seed for the random number generator, defaults to 42
    :type seed: int, optional
    :param gradient_checkpointing: Flag to indicate if gradient checkpointing should be used, defaults to True
    :type gradient_checkpointing: bool, optional
    """

    def __init__(
        self,
        pad_token_id: int,
        pad_token: str,
        model_name: str | None = None,
        actor_network: PreTrainedModelProtocol | None = None,
        model_config: dict[str, Any] | None = None,
        hp_config: HyperparameterConfig | None = None,
        index: int = 0,
        batch_size: int = 16,
        lr: float = 0.000005,
        beta: float = 0.001,
        max_grad_norm: float = 0.1,
        update_epochs: int = 1,
        calc_position_embeddings: bool = True,
        micro_batch_size_per_gpu: int | None = None,
        reduce_memory_peak: bool = False,
        device: str = "cpu",
        lora_config: LoraConfigProtocol | None = None,
        accelerator: Accelerator | None = None,
        wrap: bool = True,
        clone: bool = False,
        use_separate_reference_adapter: bool = False,
        seed: int = 42,
        gradient_checkpointing: bool = True,
    ):
        device = (
            f"cuda:{accelerator.process_index}"
            if accelerator is not None
            else ("cuda" if torch.cuda.is_available() else "cpu")
        )
        super().__init__(
            index=index,
            batch_size=batch_size,
            lr=lr,
            max_grad_norm=max_grad_norm,
            clone=clone,
            reduce_memory_peak=reduce_memory_peak,
            calc_position_embeddings=calc_position_embeddings,
            seed=seed,
            pad_token_id=pad_token_id,
            pad_token=pad_token,
            lora_config=lora_config,
            use_separate_reference_adapter=use_separate_reference_adapter,
            model_name=model_name,
            actor_network=actor_network,
            model_config=model_config,
            micro_batch_size_per_gpu=micro_batch_size_per_gpu,
            cosine_lr_schedule_config=None,
            hp_config=hp_config,
            wrap=wrap,
            device=device,
            accelerator=accelerator,
            name="DPO",
            gradient_checkpointing=gradient_checkpointing,
        )
        self.beta = beta
        self.temperature = (
            1  # Temperature for logits calculation, DPO does not use temperature
        )
        self.use_vllm = False  # DPO does not use VLLM
        self.update_epochs = update_epochs
        self._initialize_actors(actor_network, not clone)
        # Register network groups for mutations
        self.register_network_group(NetworkGroup(eval_network=self.actor, policy=True))
        if self.wrap:
            self.wrap_models()

    def get_action(
        self, obs: LLMObsType, training: bool = True
    ) -> tuple[list[torch.Tensor], list[torch.Tensor]]:
        """
        Returns the action of the agent.

        :param obs: The observation of the agent
        :type obs: LLMObsType
        :param training: Whether the agent is training or not
        :type training: bool
        :return: The action of the agent
        :rtype: tuple[list[torch.Tensor], list[torch.Tensor]]
        """
        raise NotImplementedError(
            "DPO is an offline algorithm and therefore does not require completions to be generated."
        )

    def learn(
        self,
        experiences: ExperiencesType,
        training: bool = True,
    ) -> tuple[float, float, float]:
        """
        Updates agent network parameters to learn from preference data.

        :param experiences: Batched chosen_input_ids, rejected_input_ids, chosen_attention_mask, rejected_attention_mask and rewards
        :type experiences: tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]
        :param training: Whether the agent is training or not
        :type training: bool
        :return: mean loss, mean chosen reward, mean rejected reward
        :rtype: tuple[float, float, float]
        """
        gc.collect()
        torch.cuda.empty_cache()
        # The following tensors are size [batch_size, max_length]
        chosen_input_ids: torch.Tensor = experiences["chosen_input_ids"]
        rejected_input_ids: torch.Tensor = experiences["rejected_input_ids"]
        chosen_attention_mask: torch.Tensor = experiences["chosen_attention_mask"]
        rejected_attention_mask: torch.Tensor = experiences["rejected_attention_mask"]
        # Check first that all tensors have the same max length before calculating the masks
        assert (
            chosen_input_ids.shape[1]
            == rejected_input_ids.shape[1]
            == chosen_attention_mask.shape[1]
            == rejected_attention_mask.shape[1]
        ), "All tensors must have the same max length"
        max_length = chosen_input_ids.shape[1]
        prompt_lengths: list[int] = experiences["prompt_lengths"]
        prompt_masks = LLMAlgorithm.create_prompt_masks(
            prompt_lengths, max_length=max_length
        ).to(self.device)

        # Mask has to be shifted by 1 as output log probs dims are 1 shorter than input ids as first token is used to predict the first log prob
        chosen_mask = (prompt_masks * chosen_attention_mask)[:, 1:]
        rejected_mask = (prompt_masks * rejected_attention_mask)[:, 1:]
        num_samples = chosen_input_ids.shape[0]
        batch_size = min(num_samples, self.micro_batch_size_per_gpu)
        batch_idxs = np.arange(num_samples)
        mean_loss, mean_chosen_reward, mean_rejected_reward = 0.0, 0.0, 0.0

        with torch.no_grad():
            ref_rejected_log_probs = self._get_logprobs(
                rejected_input_ids,
                batch_size,
                use_reference=True,
                eval_mode=True,
                attention_mask=rejected_attention_mask,
            )
            ref_chosen_log_probs = self._get_logprobs(
                chosen_input_ids,
                batch_size,
                use_reference=True,
                eval_mode=True,
                attention_mask=chosen_attention_mask,
            )

        for _ in range(self.update_epochs):
            for start in range(0, num_samples, batch_size):
                minibatch_idxs = batch_idxs[
                    start : min((start + batch_size), num_samples)
                ]
                (
                    batch_chosen_input_ids,
                    batch_chosen_attention_mask,
                    batch_rejected_input_ids,
                    batch_rejected_attention_mask,
                    batch_chosen_mask,
                    batch_rejected_mask,
                    batch_ref_rejected_log_probs,
                    batch_ref_chosen_log_probs,
                ) = get_experiences_samples(
                    minibatch_idxs,
                    chosen_input_ids,
                    chosen_attention_mask,
                    rejected_input_ids,
                    rejected_attention_mask,
                    chosen_mask,
                    rejected_mask,
                    ref_rejected_log_probs,
                    ref_chosen_log_probs,
                )
                batch_rejected_log_probs = self._get_logprobs(
                    batch_rejected_input_ids,
                    batch_size,
                    use_reference=False,
                    eval_mode=(not training),
                    attention_mask=batch_rejected_attention_mask,
                )
                batch_chosen_log_probs = self._get_logprobs(
                    batch_chosen_input_ids,
                    batch_size,
                    use_reference=False,
                    eval_mode=(not training),
                    attention_mask=batch_chosen_attention_mask,
                )
                loss, chosen_reward, rejected_reward = self._dpo_loss(
                    batch_chosen_log_probs,
                    batch_rejected_log_probs,
                    batch_ref_chosen_log_probs,
                    batch_ref_rejected_log_probs,
                    batch_chosen_mask,
                    batch_rejected_mask,
                )
                if training:
                    self._backward_pass(loss)
                mean_loss += loss.item()
                mean_chosen_reward += chosen_reward.mean().item()
                mean_rejected_reward += rejected_reward.mean().item()
        mean_loss /= num_samples
        mean_chosen_reward /= num_samples
        mean_rejected_reward /= num_samples
        return mean_loss, mean_chosen_reward, mean_rejected_reward

    def _dpo_loss(
        self,
        chosen_log_probs: torch.Tensor,
        rejected_log_probs: torch.Tensor,
        ref_chosen_log_probs: torch.Tensor,
        ref_rejected_log_probs: torch.Tensor,
        chosen_mask: torch.Tensor,
        rejected_mask: torch.Tensor,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Calculates the DPO loss.xw

        :param chosen_mask: Mask for the prompt and padding tokens of the chosen completions
        :type chosen_mask: torch.Tensor
        :param rejected_mask: Mask for the prompt and padding tokens of the rejected completions
        :type rejected_mask: torch.Tensor
        :param chosen_log_probs: Log probabilities of the chosen completions
        :type chosen_log_probs: torch.Tensor
        :param rejected_log_probs: Log probabilities of the rejected completions
        :type rejected_log_probs: torch.Tensor
        :param ref_chosen_log_probs: Log probabilities of the chosen completions using the reference model

        """
        # Mask and sum the logprobs
        assert (
            chosen_log_probs.shape == chosen_mask.shape
        ), f"Chosen log probabilities and mask must have the same shape, got {chosen_log_probs.shape} and {chosen_mask.shape}"
        chosen_log_probs = (chosen_log_probs * chosen_mask).sum(dim=-1)
        rejected_log_probs = (rejected_log_probs * rejected_mask).sum(dim=-1)
        ref_chosen_log_probs = (ref_chosen_log_probs * chosen_mask).sum(dim=-1)
        ref_rejected_log_probs = (ref_rejected_log_probs * rejected_mask).sum(dim=-1)
        rejected_ratio = rejected_log_probs - ref_rejected_log_probs
        chosen_ratio = chosen_log_probs - ref_chosen_log_probs
        with torch.no_grad():
            implicit_chosen_reward = self._compute_implicit_reward(
                chosen_log_probs, ref_chosen_log_probs
            )
            implicit_rejected_reward = self._compute_implicit_reward(
                rejected_log_probs, ref_rejected_log_probs
            )

        # Clean up intermediate tensors to free up memory
        chosen_log_probs = None
        rejected_log_probs = None
        ref_chosen_log_probs = None
        ref_rejected_log_probs = None
        chosen_mask = None
        rejected_mask = None

        return (
            -F.logsigmoid(self.beta * (chosen_ratio - rejected_ratio)).mean(),
            implicit_chosen_reward,
            implicit_rejected_reward,
        )

    def _compute_implicit_reward(
        self,
        log_probs: torch.Tensor,
        ref_log_probs: torch.Tensor,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Calculates the preference reward for the chosen and rejected completions.

        :param log_probs: Log probabilities of the completions
        :type log_probs: torch.Tensor
        :param ref_log_probs: Log probabilities of the completions using the reference model
        :type ref_log_probs: torch.Tensor
        :return: Implicit reward for the chosen and rejected completions
        :rtype: tuple[torch.Tensor, torch.Tensor]
        """
        implicit_reward = log_probs - ref_log_probs
        return self.beta * implicit_reward

    def test(self, env: PreferenceGym, loop: int = 1) -> torch.Tensor:
        """
        Returns the fitness (test) score tensor of the agent.

        :param env: The environment to be tested in
        :type env: PreferenceGym environment
        :param loop: Number of testing loops/episodes to complete. The returned score is the mean. Defaults to 3
        :type loop: int, optional
        :return: Test score tensor of the agent
        :rtype: torch.Tensor
        """
        with env.eval_mode(), torch.no_grad():
            prompts = env.reset()
            rewards = []
            for _ in range(loop):
                _, chosen_reward, rejected_reward = self.learn(prompts, training=False)
                reward_margin = chosen_reward - rejected_reward
                rewards.append(reward_margin)
                prompts = env.step()
        self.fitness.append(reward_margin)
        return reward_margin



import importlib

import supersuit as ss
import torch
import yaml
from accelerate import Accelerator
from pettingzoo.utils import env_logger

from agilerl.algorithms.core.registry import HyperparameterConfig, RLParameter
from agilerl.hpo.mutation import Mutations
from agilerl.hpo.tournament import TournamentSelection
from agilerl.training.train_multi_agent_on_policy import train_multi_agent_on_policy
from agilerl.utils.utils import (
    create_population,
    make_multi_agent_vect_envs,
    observation_space_channels_to_first,
    suppress_verbose_logging,
)

# !Note: If you are running this demo without having installed agilerl,
# uncomment and place the following above agilerl imports:

# import sys
# sys.path.append('../')


def main(INIT_HP, MUTATION_PARAMS, NET_CONFIG, DISTRIBUTED_TRAINING):
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    env_logger.EnvLogger.suppress_output()
    suppress_verbose_logging()

    print("============ AgileRL Multi-agent benchmarking ============")

    if DISTRIBUTED_TRAINING:
        accelerator = Accelerator()
        accelerator.wait_for_everyone()
        if accelerator.is_main_process:
            print("===== Distributed Training =====")
        accelerator.wait_for_everyone()
    else:
        accelerator = None

    print(f"DEVICE: {device}")

    def create_env(**kwargs):
        env = importlib.import_module(f"{INIT_HP['ENV_NAME']}").parallel_env(**kwargs)

        if INIT_HP["CHANNELS_LAST"]:
            # Environment processing for image based observations
            env = ss.frame_skip_v0(env, 4)
            env = ss.clip_reward_v0(env, lower_bound=-1, upper_bound=1)
            env = ss.color_reduction_v0(env, mode="B")
            env = ss.resize_v1(env, x_size=84, y_size=84)
            env = ss.frame_stack_v1(env, 4)

        return env

    env_kwargs = dict(max_cycles=25, continuous_actions=True)
    env = make_multi_agent_vect_envs(
        create_env, num_envs=INIT_HP["NUM_ENVS"], **env_kwargs
    )

    env.reset(seed=42)

    # Configure the multi-agent algo input arguments
    observation_spaces = [env.single_observation_space(agent) for agent in env.agents]
    action_spaces = [env.single_action_space(agent) for agent in env.agents]
    if INIT_HP["CHANNELS_LAST"]:
        observation_spaces = [
            observation_space_channels_to_first(obs) for obs in observation_spaces
        ]

    INIT_HP["AGENT_IDS"] = [agent_id for agent_id in env.agents]

    tournament = TournamentSelection(
        INIT_HP["TOURN_SIZE"],
        INIT_HP["ELITISM"],
        INIT_HP["POP_SIZE"],
        INIT_HP["EVAL_LOOP"],
    )

    mutations = Mutations(
        no_mutation=MUTATION_PARAMS["NO_MUT"],
        architecture=MUTATION_PARAMS["ARCH_MUT"],
        new_layer_prob=MUTATION_PARAMS["NEW_LAYER"],
        parameters=MUTATION_PARAMS["PARAMS_MUT"],
        activation=MUTATION_PARAMS["ACT_MUT"],
        rl_hp=MUTATION_PARAMS["RL_HP_MUT"],
        mutation_sd=MUTATION_PARAMS["MUT_SD"],
        rand_seed=MUTATION_PARAMS["RAND_SEED"],
        device=device,
        accelerator=accelerator,
    )

    hp_config = HyperparameterConfig(
        lr=RLParameter(min=MUTATION_PARAMS["MIN_LR"], max=MUTATION_PARAMS["MAX_LR"]),
        batch_size=RLParameter(
            min=MUTATION_PARAMS["MIN_BATCH_SIZE"],
            max=MUTATION_PARAMS["MAX_BATCH_SIZE"],
            dtype=int,
        ),
        learn_step=RLParameter(
            min=MUTATION_PARAMS["MIN_LEARN_STEP"],
            max=MUTATION_PARAMS["MAX_LEARN_STEP"],
            dtype=int,
            grow_factor=1.5,
            shrink_factor=0.75,
        ),
        ent_coef=RLParameter(
            min=MUTATION_PARAMS["MIN_ENT_COEF"],
            max=MUTATION_PARAMS["MAX_ENT_COEF"],
        ),
    )

    agent_pop = create_population(
        algo=INIT_HP["ALGO"],
        observation_space=observation_spaces,
        action_space=action_spaces,
        net_config=NET_CONFIG,
        INIT_HP=INIT_HP,
        hp_config=hp_config,
        actor_network=None,
        critic_network=None,
        population_size=INIT_HP["POP_SIZE"],
        num_envs=INIT_HP["NUM_ENVS"],
        device=device,
        accelerator=accelerator,
        torch_compiler=INIT_HP["TORCH_COMPILE"],
    )

    try:
        train_multi_agent_on_policy(
            env,
            INIT_HP["ENV_NAME"],
            INIT_HP["ALGO"],
            agent_pop,
            sum_scores=True,
            INIT_HP=INIT_HP,
            MUT_P=MUTATION_PARAMS,
            swap_channels=INIT_HP["CHANNELS_LAST"],
            max_steps=INIT_HP["MAX_STEPS"],
            evo_steps=INIT_HP["EVO_STEPS"],
            eval_steps=INIT_HP["EVAL_STEPS"],
            eval_loop=INIT_HP["EVAL_LOOP"],
            target=INIT_HP["TARGET_SCORE"],
            tournament=tournament,
            mutation=mutations,
            wb=INIT_HP["WANDB"],
            accelerator=accelerator,
        )
    except Exception as e:
        env.close()
        raise e

    if str(device) == "cuda":
        torch.cuda.empty_cache()


if __name__ == "__main__":
    config = "configs/training/multi_agent/ippo.yaml"
    with open(config) as file:
        config = yaml.safe_load(file)
    INIT_HP = config["INIT_HP"]
    MUTATION_PARAMS = config["MUTATION_PARAMS"]
    NET_CONFIG = config["NET_CONFIG"]
    DISTRIBUTED_TRAINING = config["DISTRIBUTED_TRAINING"]
    main(INIT_HP, MUTATION_PARAMS, NET_CONFIG, DISTRIBUTED_TRAINING)



from unittest.mock import Mock, patch

import numpy as np
import pytest
import torch
import torch.nn as nn
from gymnasium import spaces
from torch._dynamo.eval_frame import OptimizedModule

from agilerl.modules.cnn import MutableKernelSizes
from agilerl.utils.algo_utils import (
    get_input_size_from_space,
    get_output_size_from_space,
)
from agilerl.utils.evolvable_networks import (
    compile_model,
    create_cnn,
    create_mlp,
    get_activation,
)


######### Test get_activation #########
def test_returns_correct_activation_function_for_all_supported_names():
    activation_names = [
        "Tanh",
        "Identity",
        "ReLU",
        "ELU",
        "Softsign",
        "Sigmoid",
        "GumbelSoftmax",
        "Softplus",
        "Softmax",
        "LeakyReLU",
        "PReLU",
        "GELU",
    ]
    for name in activation_names:
        activation = get_activation(name)
        assert isinstance(activation, nn.Module)


@pytest.mark.parametrize(
    "input_shape, channel_size, kernel_size, stride_size",
    [
        ([1, 16, 16], [32, 16], [3, 2], [1, 1]),
    ],
)
def test_calc_max_kernel_sizes(input_shape, channel_size, kernel_size, stride_size):
    mutable_kernel_sizes = MutableKernelSizes(
        sizes=kernel_size,
        cnn_block_type="Conv2d",
        sample_input=torch.randn(1, 3, 32, 32),
        rng=np.random.default_rng(42),
    )
    max_kernel_sizes = mutable_kernel_sizes.calc_max_kernel_sizes(
        channel_size, stride_size, input_shape
    )
    assert max_kernel_sizes == [3, 3]


@pytest.mark.parametrize(
    "input_shape, channel_size, kernel_size, stride_size",
    [
        ([1, 3, 3], [32, 16], [1, 1], [1, 1]),
    ],
)
def test_max_kernel_size_negative(input_shape, channel_size, kernel_size, stride_size):
    mutable_kernel_sizes = MutableKernelSizes(
        sizes=kernel_size,
        cnn_block_type="Conv2d",
        sample_input=torch.randn(1, 3, 32, 32),
        rng=np.random.default_rng(42),
    )
    max_kernel_sizes = mutable_kernel_sizes.calc_max_kernel_sizes(
        channel_size, stride_size, input_shape
    )
    assert max_kernel_sizes == [1, 1]


@pytest.mark.parametrize("output_vanish, noisy", [(True, True), (False, True)])
def test_create_mlp(output_vanish, noisy):
    net = create_mlp(
        10,
        4,
        [32, 32],
        output_vanish=output_vanish,
        output_activation=None,
        noisy=noisy,
    )
    assert isinstance(net, nn.Module)


######### Test create_mlp and create_cnn########
@pytest.mark.parametrize("noisy, output_vanish", [(False, True), (True, False)])
def test_create_cnn(noisy, output_vanish):
    feature_net = create_cnn(
        "Conv2d",
        1,
        [32, 32],
        [3, 3],
        [1, 1],
        "feature",
        layer_norm=True,
    )

    head = create_mlp(
        10,
        4,
        [64, 64],
        output_activation=None,
        noisy=noisy,
        name="value",
        output_vanish=output_vanish,
    )

    feature_net = nn.ModuleDict(feature_net)

    assert isinstance(head, nn.Module)
    assert isinstance(feature_net, nn.Module)


def test_compile_model():
    model = nn.Linear(10, 10)

    # Test with mocked torch.compile
    with patch("torch.compile", return_value=model):
        compiled_model = compile_model(model, mode="default")
        # Should have called torch.compile
        assert compiled_model is model

    # Test with already compiled model (mock OptimizedModule)
    with patch("torch._dynamo.eval_frame.OptimizedModule", type(OptimizedModule)):
        optimized_model = Mock(spec=OptimizedModule)

        # Should return the model without recompiling
        result = compile_model(optimized_model, mode="default")
        assert result is optimized_model

    # Test with mode=None
    result = compile_model(model, mode=None)
    assert result is model


@pytest.mark.parametrize("noisy, output_vanish", [(False, True), (True, False)])
def test_create_cnn_multi(noisy, output_vanish):
    feature_net = create_cnn(
        "Conv3d",
        1,
        [32, 32],
        [3, 3],
        [1, 1],
        "feature",
        layer_norm=True,
    )
    head = create_mlp(
        10,
        4,
        [64, 64],
        output_activation=None,
        noisy=noisy,
        name="value",
        output_vanish=output_vanish,
    )

    feature_net = nn.ModuleDict(feature_net)

    assert isinstance(head, nn.Module)
    assert isinstance(feature_net, nn.Module)


######### Test get_input_size_from_space #########
class TestGetInputSizeFromSpace:
    """Comprehensive tests for get_input_size_from_space function."""

    def test_discrete_space(self):
        """Test Discrete space returns correct tuple."""
        space = spaces.Discrete(5)
        result = get_input_size_from_space(space)
        assert result == (5,)

    def test_multi_discrete_space(self):
        """Test MultiDiscrete space returns sum of nvec."""
        space = spaces.MultiDiscrete([2, 3, 4])
        result = get_input_size_from_space(space)
        assert result == (9,)  # 2 + 3 + 4

    def test_box_space_1d(self):
        """Test 1D Box space returns shape."""
        space = spaces.Box(low=-1, high=1, shape=(4,), dtype=np.float32)
        result = get_input_size_from_space(space)
        assert result == (4,)

    def test_box_space_2d(self):
        """Test 2D Box space returns shape."""
        space = spaces.Box(low=-1, high=1, shape=(3, 4), dtype=np.float32)
        result = get_input_size_from_space(space)
        assert result == (3, 4)

    def test_box_space_3d_image(self):
        """Test 3D Box space (image) returns shape."""
        space = spaces.Box(low=0, high=255, shape=(3, 32, 32), dtype=np.uint8)
        result = get_input_size_from_space(space)
        assert result == (3, 32, 32)

    def test_multi_binary_space(self):
        """Test MultiBinary space returns tuple of n."""
        space = spaces.MultiBinary(8)
        result = get_input_size_from_space(space)
        assert result == (8,)

    def test_dict_space(self):
        """Test Dict space returns dict of sizes."""
        space = spaces.Dict(
            {
                "position": spaces.Box(low=-10, high=10, shape=(2,), dtype=np.float32),
                "velocity": spaces.Discrete(4),
                "sensor": spaces.MultiBinary(6),
            }
        )
        result = get_input_size_from_space(space)
        expected = {"position": (2,), "velocity": (4,), "sensor": (6,)}
        assert result == expected

    def test_tuple_space(self):
        """Test Tuple space returns tuple of sizes."""
        space = spaces.Tuple(
            (
                spaces.Box(low=-1, high=1, shape=(3,), dtype=np.float32),
                spaces.Discrete(5),
                spaces.MultiDiscrete([2, 3]),
            )
        )
        result = get_input_size_from_space(space)
        expected = ((3,), (5,), (5,))  # (3,), (5,), (2+3,)
        assert result == expected

    def test_list_of_spaces(self):
        """Test list of spaces returns tuple of sizes."""
        spaces_list = [
            spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32),
            spaces.Discrete(3),
        ]
        result = get_input_size_from_space(spaces_list)
        expected = ((2,), (3,))
        assert result == expected

    def test_nested_dict_space(self):
        """Test nested Dict space."""
        space = spaces.Dict(
            {
                "observations": spaces.Dict(
                    {
                        "position": spaces.Box(
                            low=-10, high=10, shape=(3,), dtype=np.float32
                        ),
                        "velocity": spaces.Box(
                            low=-1, high=1, shape=(3,), dtype=np.float32
                        ),
                    }
                ),
                "action_mask": spaces.MultiBinary(5),
            }
        )
        result = get_input_size_from_space(space)
        expected = {
            "observations": {"position": (3,), "velocity": (3,)},
            "action_mask": (5,),
        }
        assert result == expected

    def test_nested_tuple_space(self):
        """Test nested Tuple space."""
        space = spaces.Tuple(
            (
                spaces.Tuple(
                    (
                        spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32),
                        spaces.Discrete(3),
                    )
                ),
                spaces.Box(low=0, high=1, shape=(4,), dtype=np.float32),
            )
        )
        result = get_input_size_from_space(space)
        expected = (((2,), (3,)), (4,))
        assert result == expected

    def test_unsupported_space_raises_error(self):
        """Test that unsupported space types raise AttributeError."""

        # Create a mock space that's not supported
        class UnsupportedSpace:
            pass

        unsupported_space = UnsupportedSpace()
        with pytest.raises(AttributeError, match="Can't access state dimensions"):
            get_input_size_from_space(unsupported_space)

    def test_empty_dict_space(self):
        """Test empty Dict space."""
        space = spaces.Dict({})
        result = get_input_size_from_space(space)
        assert result == {}

    def test_empty_tuple_space(self):
        """Test empty Tuple space."""
        space = spaces.Tuple(())
        result = get_input_size_from_space(space)
        assert result == ()


######### Test get_output_size_from_space #########
class TestGetOutputSizeFromSpace:
    """Comprehensive tests for get_output_size_from_space function."""

    def test_discrete_space(self):
        """Test Discrete space returns n."""
        space = spaces.Discrete(5)
        result = get_output_size_from_space(space)
        assert result == 5

    def test_multi_discrete_space(self):
        """Test MultiDiscrete space returns sum of nvec."""
        space = spaces.MultiDiscrete([2, 3, 4])
        result = get_output_size_from_space(space)
        assert result == 9  # 2 + 3 + 4

    def test_box_space_1d(self):
        """Test 1D Box space returns first dimension."""
        space = spaces.Box(low=-1, high=1, shape=(4,), dtype=np.float32)
        result = get_output_size_from_space(space)
        assert result == 4

    def test_box_space_multi_dim(self):
        """Test multi-dimensional Box space returns first dimension."""
        space = spaces.Box(low=-1, high=1, shape=(3, 4, 5), dtype=np.float32)
        result = get_output_size_from_space(space)
        assert result == 3

    def test_multi_binary_space(self):
        """Test MultiBinary space returns n."""
        space = spaces.MultiBinary(8)
        result = get_output_size_from_space(space)
        assert result == 8

    def test_dict_space(self):
        """Test Dict space returns dict of sizes."""
        space = spaces.Dict(
            {
                "discrete_action": spaces.Discrete(4),
                "continuous_action": spaces.Box(
                    low=-1, high=1, shape=(2,), dtype=np.float32
                ),
                "binary_action": spaces.MultiBinary(3),
            }
        )
        result = get_output_size_from_space(space)
        expected = {"discrete_action": 4, "continuous_action": 2, "binary_action": 3}
        assert result == expected

    def test_list_of_spaces(self):
        """Test list of spaces returns tuple of sizes."""
        spaces_list = [
            spaces.Discrete(3),
            spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32),
        ]
        result = get_output_size_from_space(spaces_list)
        expected = (3, 2)
        assert result == expected

    def test_nested_dict_space(self):
        """Test nested Dict space."""
        space = spaces.Dict(
            {
                "movement": spaces.Dict(
                    {
                        "direction": spaces.Discrete(4),
                        "speed": spaces.Box(
                            low=0, high=1, shape=(1,), dtype=np.float32
                        ),
                    }
                ),
                "special_action": spaces.MultiBinary(2),
            }
        )
        result = get_output_size_from_space(space)
        expected = {"movement": {"direction": 4, "speed": 1}, "special_action": 2}
        assert result == expected

    def test_unsupported_space_raises_error(self):
        """Test that unsupported space types raise AttributeError."""

        # Create a mock space that's not supported
        class UnsupportedSpace:
            pass

        unsupported_space = UnsupportedSpace()
        with pytest.raises(AttributeError, match="Can't access action dimensions"):
            get_output_size_from_space(unsupported_space)

    def test_empty_dict_space(self):
        """Test empty Dict space."""
        space = spaces.Dict({})
        result = get_output_size_from_space(space)
        assert result == {}

    def test_box_space_scalar(self):
        """Test scalar Box space (shape=())."""
        space = spaces.Box(low=-1, high=1, shape=(), dtype=np.float32)
        # This should raise IndexError since shape[0] doesn't exist
        with pytest.raises(IndexError):
            get_output_size_from_space(space)

    def test_python_tuple(self):
        """Test that Python tuples work."""
        # Python tuple should work
        python_tuple = (
            spaces.Discrete(3),
            spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32),
        )
        result = get_output_size_from_space(python_tuple)
        assert result == (3, 2)


######### Test Edge Cases and Complex Scenarios #########
class TestSpaceSizeEdgeCases:
    """Test edge cases and complex scenarios for both functions."""

    def test_complex_mixed_nested_structure(self):
        """Test complex nested structure with mixed space types."""
        # Complex observation space
        obs_space = spaces.Dict(
            {
                "visual": spaces.Tuple(
                    (
                        spaces.Box(low=0, high=255, shape=(3, 64, 64), dtype=np.uint8),
                        spaces.Box(low=0, high=255, shape=(1, 32, 32), dtype=np.uint8),
                    )
                ),
                "proprioception": spaces.Dict(
                    {
                        "position": spaces.Box(
                            low=-10, high=10, shape=(6,), dtype=np.float32
                        ),
                        "velocity": spaces.Box(
                            low=-5, high=5, shape=(6,), dtype=np.float32
                        ),
                        "discrete_state": spaces.Discrete(10),
                    }
                ),
                "sensors": spaces.Tuple(
                    (spaces.MultiBinary(16), spaces.MultiDiscrete([4, 8, 2]))
                ),
            }
        )

        # Complex action space - Using Python tuple instead of spaces.Tuple
        # since get_output_size_from_space doesn't support spaces.Tuple
        action_space = spaces.Dict(
            {
                "motor_commands": spaces.Box(
                    low=-1, high=1, shape=(12,), dtype=np.float32
                ),
                "binary_switches": spaces.MultiBinary(8),
            }
        )

        # Test input size
        input_result = get_input_size_from_space(obs_space)
        expected_input = {
            "visual": ((3, 64, 64), (1, 32, 32)),
            "proprioception": {
                "position": (6,),
                "velocity": (6,),
                "discrete_state": (10,),
            },
            "sensors": ((16,), (14,)),
        }
        assert input_result == expected_input

        # Test output size
        output_result = get_output_size_from_space(action_space)
        expected_output = {
            "motor_commands": 12,
            "binary_switches": 8,
        }
        assert output_result == expected_output

    def test_consistency_with_gymnasium_spaces(self):
        """Test that results are consistent with actual gymnasium space sampling."""
        spaces_to_test = [
            spaces.Discrete(7),
            spaces.Box(low=-1, high=1, shape=(5,), dtype=np.float32),
            spaces.MultiDiscrete([3, 4, 5]),
            spaces.MultiBinary(10),
            spaces.Box(low=0, high=255, shape=(3, 28, 28), dtype=np.uint8),
        ]

        for space in spaces_to_test:
            # Get the size from our function
            input_size = get_input_size_from_space(space)

            if isinstance(space, spaces.Discrete):
                # For discrete spaces, we expect tuple with n values
                assert input_size == (space.n,)
            elif isinstance(space, spaces.Box):
                # For box spaces, should match the shape
                assert input_size == space.shape
            elif isinstance(space, spaces.MultiDiscrete):
                # Should be sum of nvec
                assert input_size == (sum(space.nvec),)
            elif isinstance(space, spaces.MultiBinary):
                # Should be n
                assert input_size == (space.n,)

    def test_python_dict_vs_gymnasium_dict(self):
        """Test that function works with both Python dict and gymnasium Dict."""
        # Create equivalent spaces
        gym_dict_space = spaces.Dict(
            {
                "a": spaces.Discrete(3),
                "b": spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32),
            }
        )

        python_dict_space = {
            "a": spaces.Discrete(3),
            "b": spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32),
        }

        # Both should give same result
        gym_result = get_input_size_from_space(gym_dict_space)
        python_result = get_input_size_from_space(python_dict_space)

        expected = {"a": (3,), "b": (2,)}
        assert gym_result == expected
        assert python_result == expected

    def test_tuple_vs_list_vs_gymnasium_tuple(self):
        """Test that function works with tuple, list, and gymnasium Tuple."""
        discrete_space = spaces.Discrete(3)
        box_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)

        # Create equivalent structures
        python_tuple = (discrete_space, box_space)
        python_list = [discrete_space, box_space]
        gym_tuple = spaces.Tuple((discrete_space, box_space))

        # All should give same result
        tuple_result = get_input_size_from_space(python_tuple)
        list_result = get_input_size_from_space(python_list)
        gym_result = get_input_size_from_space(gym_tuple)

        expected = ((3,), (2,))
        assert tuple_result == expected
        assert list_result == expected
        assert gym_result == expected



import torch
import yaml

from agilerl.algorithms.core.registry import HyperparameterConfig, RLParameter
from agilerl.hpo.mutation import Mutations
from agilerl.hpo.tournament import TournamentSelection
from agilerl.training.train_on_policy import train_on_policy
from agilerl.utils.utils import (
    create_population,
    make_vect_envs,
    observation_space_channels_to_first,
    print_hyperparams,
)

# !Note: If you are running this demo without having installed agilerl,
# uncomment and place the following above agilerl imports:

# import sys
# sys.path.append('../')


def main(INIT_HP, MUTATION_PARAMS, NET_CONFIG, use_net=False):
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    print("============ AgileRL ============")
    print(f"DEVICE: {device}")

    env = make_vect_envs(INIT_HP["ENV_NAME"], num_envs=INIT_HP["NUM_ENVS"])

    observation_space = env.single_observation_space
    action_space = env.single_action_space
    if INIT_HP["CHANNELS_LAST"]:
        observation_space = observation_space_channels_to_first(observation_space)

    tournament = TournamentSelection(
        INIT_HP["TOURN_SIZE"],
        INIT_HP["ELITISM"],
        INIT_HP["POP_SIZE"],
        INIT_HP["EVAL_LOOP"],
    )
    mutations = Mutations(
        no_mutation=MUTATION_PARAMS["NO_MUT"],
        architecture=MUTATION_PARAMS["ARCH_MUT"],
        new_layer_prob=MUTATION_PARAMS["NEW_LAYER"],
        parameters=MUTATION_PARAMS["PARAMS_MUT"],
        activation=MUTATION_PARAMS["ACT_MUT"],
        rl_hp=MUTATION_PARAMS["RL_HP_MUT"],
        mutation_sd=MUTATION_PARAMS["MUT_SD"],
        rand_seed=MUTATION_PARAMS["RAND_SEED"],
        device=device,
    )

    hp_config = HyperparameterConfig(
        lr=RLParameter(min=MUTATION_PARAMS["MIN_LR"], max=MUTATION_PARAMS["MAX_LR"]),
        batch_size=RLParameter(
            min=MUTATION_PARAMS["MIN_BATCH_SIZE"],
            max=MUTATION_PARAMS["MAX_BATCH_SIZE"],
            dtype=int,
        ),
    )

    NET_CONFIG["encoder_config"]["device"] = device

    agent_pop = create_population(
        algo=INIT_HP["ALGO"],
        observation_space=observation_space,
        action_space=action_space,
        net_config=NET_CONFIG,
        INIT_HP=INIT_HP,
        hp_config=hp_config,
        population_size=INIT_HP["POP_SIZE"],
        num_envs=INIT_HP["NUM_ENVS"],
        device=device,
    )

    trained_pop, pop_fitnesses = train_on_policy(
        env,
        INIT_HP["ENV_NAME"],
        INIT_HP["ALGO"],
        agent_pop,
        INIT_HP=INIT_HP,
        MUT_P=MUTATION_PARAMS,
        swap_channels=INIT_HP["CHANNELS_LAST"],
        max_steps=INIT_HP["MAX_STEPS"],
        evo_steps=INIT_HP["EVO_STEPS"],
        eval_steps=INIT_HP["EVAL_STEPS"],
        eval_loop=INIT_HP["EVAL_LOOP"],
        target=INIT_HP["TARGET_SCORE"],
        tournament=tournament,
        mutation=mutations,
        wb=INIT_HP["WANDB"],
    )

    print_hyperparams(trained_pop)

    if str(device) == "cuda":
        torch.cuda.empty_cache()

    env.close()


if __name__ == "__main__":
    with open("configs/training/ppo/ppo_resnet.yaml") as file:
        ppo_config = yaml.safe_load(file)
    INIT_HP = ppo_config["INIT_HP"]
    MUTATION_PARAMS = ppo_config["MUTATION_PARAMS"]
    NET_CONFIG = ppo_config["NET_CONFIG"]
    main(INIT_HP, MUTATION_PARAMS, NET_CONFIG)
